from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
import pytest

setup_logging()

@pytest.mark.parametrize('model', [Zheng07Model, Hearin15Model, Leauthaud11Model])
def test_consistency(model):

    # correct answers
    truth = {}
    truth['cosmology'] = cosmology.Planck15.to_astropy()
    truth['redshift'] = 0.55
    truth['prim_haloprop_key'] = 'halo_m200m'
    truth['halo_boundary_key'] = 'halo_r200m'

    # make the model
    model = model.to_halotools(truth['cosmology'], truth['redshift'], '200m')

    # test consistency of attributes
    attrs = ['cosmology', 'redshift', 'prim_haloprop_key', 'halo_boundary_key']
    for a in attrs:
        vals = []
        for m in model.model_dictionary:
            comp = model.model_dictionary[m]
            y = getattr(comp, a, None)
            if y is not None: vals.append(y)

        assert all(v == truth[a] for v in vals)



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging
from nbodykit.transform import ConstantArray
import pytest

# debug logging
setup_logging("debug")

@MPITest([1, 4])
def test_sky_to_cartesian(comm):

    cosmo = cosmology.Planck15

    # make source
    s = RandomCatalog(csize=100, seed=42, comm=comm)

    # ra, dec, z
    s['z']   = s.rng.normal(loc=0.5, scale=0.1)
    s['ra']  = s.rng.uniform(low=110, high=260)
    s['dec'] = s.rng.uniform(low=-3.6, high=60)

    # make the position array
    s['Position1'] = transform.SkyToCartesian(s['ra'], s['dec'], s['z'], cosmo)

    # wrong name
    with pytest.warns(FutureWarning):
        s['Position0'] = transform.SkyToCartesion(s['ra'], s['dec'], s['z'], cosmo)

    # requires dask array
    with pytest.raises(TypeError):
        s['Position1'] = transform.SkyToCartesian(s['ra'].compute(), s['dec'], s['z'], cosmo)

@MPITest([1, 4])
def test_cartesian_to_equatorial(comm):

    # make source
    s = UniformCatalog(nbar=10000, BoxSize=1.0, comm=comm)

    # get RA, DEC
    ra, dec = transform.CartesianToEquatorial(s['Position'], observer=[0.5, 0.5, 0.5])

    # check bounds
    assert ((ra >= 0.)&(ra < 360.)).all().compute()
    assert ((dec >= -90)&(dec < 90.)).all().compute()

@MPITest([1, 4])
def test_cartesian_to_sky(comm):
    cosmo = cosmology.Planck15

    # make source
    s = UniformCatalog(nbar=10000, BoxSize=1.0, seed=42, comm=comm)

    # get RA, DEC, Z
    ra, dec, z = transform.CartesianToSky(s['Position'], cosmo)

    # pos needs to be a dask array
    with pytest.raises(TypeError):
        _ = transform.CartesianToSky(s['Position'].compute(), cosmo)

    # reverse and check
    pos2 = transform.SkyToCartesian(ra, dec, z, cosmo)
    numpy.testing.assert_allclose(s['Position'], pos2, rtol=1e-5)

@MPITest([1, 4])
def test_cartesian_to_sky_velocity(comm):
    cosmo = cosmology.Planck15

    # make source
    s = UniformCatalog(nbar=1e-5, BoxSize=1380., seed=42, comm=comm)

    # real-space redshift
    _, _, z_real = transform.CartesianToSky(s['Position'], cosmo,
                                            observer=[-1e3, -1e3, -1e3])
    # redshift-space redshift
    _, _, z_redshift = transform.CartesianToSky(s['Position'], cosmo,
                                                velocity=s['Velocity'],
                                                observer=[-1e3, -1e3, -1e3])

    numpy.testing.assert_allclose(z_real, z_redshift, rtol=1e-3)

    # bad z max value
    with pytest.raises(ValueError):
        _, _, z = transform.CartesianToSky(s['Position'], cosmo, observer=[-1e4, -1e4, -1e4], zmax=0.5)
        z = z.compute()



@MPITest([1, 4])
def test_stack_columns(comm):

    # make source
    s = RandomCatalog(csize=100, seed=42, comm=comm)

    # add x,y,z
    s['x'] = s.rng.uniform(0, 2600.)
    s['y'] = s.rng.uniform(0, 2600.)
    s['z'] = s.rng.uniform(0, 2600.)

    # stack
    s['Position'] = transform.StackColumns(s['x'], s['y'], s['z'])

    # test equality
    x, y, z = s.compute(s['x'], s['y'], s['z'])
    pos = numpy.vstack([x,y,z]).T
    numpy.testing.assert_array_equal(pos, s['Position'])

    # requires dask array
    with pytest.raises(TypeError):
        s['Position'] = transform.StackColumns(x,y,z)

@MPITest([1, 4])
def test_combine(comm):

    # make two sources
    s1 = UniformCatalog(3e-6, 2600, comm=comm)
    s2 = UniformCatalog(3e-6, 2600, comm=comm)

    # concatenate all columns
    cat = transform.ConcatenateSources(s1, s2)

    # check the size and columns
    assert cat.size == s1.size + s2.size
    assert set(cat.columns) == set(s1.columns)

    # only one column
    cat = transform.ConcatenateSources(s1, s2, columns='Position')
    pos = numpy.concatenate([s1['Position'], s2['Position']], axis=0)
    numpy.testing.assert_array_equal(pos, cat['Position'])

    # fail on invalid column
    with pytest.raises(ValueError):
        cat = transform.ConcatenateSources(s1, s2, columns='InvalidColumn')

@MPITest([1])
def test_constarray(comm):
    a = ConstantArray(1.0, 1, chunks=1000)
    assert len(a) == 1
    print(a.shape)
    assert a.shape == (1,)
    a = ConstantArray([1.0, 1.0], 1, chunks=1000)
    assert a.shape == (1, 2)

    a = ConstantArray([1.0, 1.0], 3, chunks=1000)
    assert a.shape == (3, 2)



from .binary import BinaryFile

class TPMBinaryFile(BinaryFile):
    """
    Read snapshot binary files from Martin White's TPM simulations.

    These files are stored column-wise with a format, with a
    header of size 28 bytes to begin the file.

    The columns are:

    * Position : 'f4', 'f8' precision
        the position data
    * Velocity : 'f4', 'f8' precision
        the velocity data
    * ID : 'u8' precision
        integers specfiying the particle ID

    Parameters
    ----------
    path : str
        the path to the binary file to load
    precision : {'f4', 'f8'}, optional
        the string dtype specifying the precision

    References
    ----------
    White M., 2002, ApJS, 579, 16
    """
    def __init__(self, path, precision='f4'):

        if precision not in ['f4', 'f8']:
            raise ValueError("precision should be either 'f4' or 'f8'")

        dtype = [('Position', (precision, 3)), ('Velocity', (precision, 3)), ('ID', 'u8')]
        BinaryFile.__init__(self, path, dtype=dtype, header_size=28)



from nbodykit.cosmology import Cosmology

class HODModel(object):
    """
    A class providing an interface beteween HOD models in
    :mod:`nbodykit` and :mod:`halotools`.

    See the static method :func:`to_halotools` to create a specified
    :mod:`halotools` HOD model given:

    - cosmology
    - redshift
    - mass definition

    .. note::
        Here, mass definition is used to place satellites in halos using
        a NFW profile in order to convert mass to radius.
    """
    @staticmethod
    def to_halotools(cosmo, redshift, mdef, concentration_key=None, **kwargs):
        """
        Return a {func_name} HOD model in the format of :mod:`halotools`.

        See :func:`halotools.empirical_models.{name}_model_dictionary`.

        Parameters
        ----------
        cosmo :
            the nbodykit or astropy Cosmology object to use in the model
        redshift : float
            the desired redshift of the model
        mdef : str, optional
            string specifying mass definition, used for computing default
            halo radii and concentration; should be 'vir' or 'XXXc' or
            'XXXm' where 'XXX' is an int specifying the overdensity
        concentration_key : str
            the name of the column that will specify concentration; if not
            provided, the analytic formula from
            `Dutton and Maccio 2014 <https://arxiv.org/abs/1402.7073>`_
            is used.
        **kwargs :
            additional keywords passed to the model components; see the
            Halotools documentation for further details

        Returns
        -------
        :class:`~halotools.empirical_models.HodModelFactory`
            the halotools object implementing the HOD model
        """
        raise NotImplementedError

# NOTE: we are making zheng 07 separately due to astropy/halotools#827
class Zheng07Model(HODModel):

    @staticmethod
    def to_halotools(cosmo, redshift, mdef, concentration_key=None, **kwargs):
        """
        Return the Zheng 07 HOD model.

        See :func:`halotools.empirical_models.zheng07_model_dictionary`.

        Parameters
        ----------
        cosmo :
            the nbodykit or astropy Cosmology object to use in the model
        redshift : float
            the desired redshift of the model
        mdef : str, optional
            string specifying mass definition, used for computing default
            halo radii and concentration; should be 'vir' or 'XXXc' or
            'XXXm' where 'XXX' is an int specifying the overdensity
        concentration_key : str
            the name of the column that will specify concentration; if not
            provided, the analytic formula from
            `Dutton and Maccio 2014 <https://arxiv.org/abs/1402.7073>`_
            is used.
        **kwargs :
            additional keywords passed to the model components; see the
            Halotools documentation for further details

        Returns
        -------
        :class:`~halotools.empirical_models.HodModelFactory`
            the halotools object implementing the HOD model
        """
        from halotools.empirical_models import Zheng07Sats, Zheng07Cens, NFWPhaseSpace, TrivialPhaseSpace
        from halotools.empirical_models import HodModelFactory

        kwargs.setdefault('modulate_with_cenocc', True)

        # need astropy Cosmology
        if isinstance(cosmo, Cosmology):
            cosmo = cosmo.to_astropy()

        # determine concentration key
        if concentration_key is None:
            conc_mass_model = 'dutton_maccio14'
        else:
            conc_mass_model = 'direct_from_halo_catalog'

        # determine mass column
        mass_key = 'halo_m' + mdef

        # occupation functions
        cenocc = Zheng07Cens(prim_haloprop_key=mass_key, **kwargs)
        satocc = Zheng07Sats(prim_haloprop_key=mass_key, cenocc_model=cenocc, **kwargs)
        satocc._suppress_repeated_param_warning = True

        # profile functions
        kwargs.update({'cosmology':cosmo, 'redshift':redshift, 'mdef':mdef})
        censprof = TrivialPhaseSpace(**kwargs)
        satsprof = NFWPhaseSpace(conc_mass_model=conc_mass_model, **kwargs)

        # make the model
        model = {}
        model['centrals_occupation'] = cenocc
        model['centrals_profile'] = censprof
        model['satellites_occupation'] = satocc
        model['satellites_profile'] = satsprof
        return HodModelFactory(**model)

def HODModelFactory(name, func_name):
    """
    Factory to generate the functions that will return one of the
    pre-built Halotools HOD models.

    Parameters
    ----------
    name : str
        a name of a pre-built HOD model name from Halotools

    Returns
    -------
    callable :
        a function that can compute the specified Halotools model from
        the input cosmology, redshift, mass definition, etc.
    """
    import textwrap

    def to_halotools(cosmo, redshift, mdef, concentration_key=None, **kwargs):
        """
        Parameters
        ----------
        cosmo :
            the nbodykit or astropy Cosmology object to use in the model
        redshift : float
            the desired redshift of the model
        mdef : str, optional
            string specifying mass definition, used for computing default
            halo radii and concentration; should be 'vir' or 'XXXc' or
            'XXXm' where 'XXX' is an int specifying the overdensity
        concentration_key : str
            the name of the column that will specify concentration; if not
            provided, the analytic formula from
            `Dutton and Maccio 2014 <https://arxiv.org/abs/1402.7073>`_
            is used.
        **kwargs :
            additional keywords passed to the model components; see the
            Halotools documentation for further details

        Returns
        -------
        :class:`~halotools.empirical_models.HodModelFactory`
            the halotools object implementing the HOD model
        """
        from halotools.empirical_models import PrebuiltHodModelFactory

        # modulate sats with cen occ model by default
        kwargs.setdefault('modulate_with_cenocc', True)

        # need astropy Cosmology
        if isinstance(cosmo, Cosmology):
            cosmo = cosmo.to_astropy()

        # determine concentration key
        if concentration_key is None:
            conc_mass_model = 'dutton_maccio14'
        else:
            conc_mass_model = 'direct_from_halo_catalog'

        # determine mass column
        mass_key = 'halo_m' + mdef

        # the configuration
        kwargs.update({'cosmology':cosmo, 'redshift':redshift, 'mdef':mdef, 'prim_haloprop_key':mass_key})
        return PrebuiltHodModelFactory(name, **kwargs)

    # make the doc string
    hdr = "Return the {func_name} HOD model from Halotools.\n\n".format(func_name=func_name)
    hdr += "See :func:`halotools.empirical_models.{func_name}_model_dictionary` for further details.\n"
    to_halotools.__doc__ = hdr + textwrap.dedent(to_halotools.__doc__)

    # make the new class object and return it
    newclass = type(name, (HODModel,),{"to_halotools": staticmethod(to_halotools), "__doc__":__doc__})
    return newclass

# generate the specialized models
Leauthaud11Model = HODModelFactory('leauthaud11', 'Leauthaud11Model')
Hearin15Model    = HODModelFactory('hearin15', 'Hearin15Model')

__all__ = ['Zheng07Model', 'Leauthaud11Model', 'Hearin15Model']



import types
import six
import inspect
import re
from warnings import warn


__version__ = '0.2.3'

__author__ = 'Philipp Sommer'


try:
    from matplotlib.cbook import dedent as dedents
except ImportError:
    from textwrap import dedent as _dedents

    def dedents(s):
        return '\n'.join(_dedents(s or '').splitlines()[1:])


substitution_pattern = re.compile(
    r"""(?<!%)(%%)*%(?!%)   # uneven number of %
        \((?P<key>(?s).*?)\)# key enclosed in brackets""", re.VERBOSE)


summary_patt = re.compile(r'(?s).*?(?=(\n\s*\n)|$)')


class _StrWithIndentation(object):
    """A convenience class that indents the given string if requested through
    the __str__ method"""

    def __init__(self, s, indent=0, *args, **kwargs):
        self._indent = '\n' + ' ' * indent
        self._s = s

    def __str__(self):
        return self._indent.join(self._s.splitlines())

    def __repr__(self):
        return repr(self._indent.join(self._s.splitlines()))


def safe_modulo(s, meta, checked='', print_warning=True, stacklevel=2):
    """Safe version of the modulo operation (%) of strings

    Parameters
    ----------
    s: str
        string to apply the modulo operation with
    meta: dict or tuple
        meta informations to insert (usually via ``s % meta``)
    checked: {'KEY', 'VALUE'}, optional
        Security parameter for the recursive structure of this function. It can
        be set to 'VALUE' if an error shall be raised when facing a TypeError
        or ValueError or to 'KEY' if an error shall be raised when facing a
        KeyError. This parameter is mainly for internal processes.
    print_warning: bool
        If True and a key is not existent in `s`, a warning is raised
    stacklevel: int
        The stacklevel for the :func:`warnings.warn` function


    Examples
    --------
    The effects are demonstrated by this example::

        >>> from docrep import safe_modulo
        >>> s = "That's %(one)s string %(with)s missing 'with' and %s key"
        >>> s % {'one': 1}          # raises KeyError because of missing 'with'
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        KeyError: 'with'
        >>> s % {'one': 1, 'with': 2}        # raises TypeError because of '%s'
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        TypeError: not enough arguments for format string
        >>> safe_modulo(s, {'one': 1})
        "That's 1 string %(with)s missing 'with' and %s key"
    """
    try:
        return s % meta
    except (ValueError, TypeError, KeyError):
        # replace the missing fields by %%
        keys = substitution_pattern.finditer(s)
        for m in keys:
            key = m.group('key')
            if not isinstance(meta, dict) or key not in meta:
                if print_warning:
                    warn("%r is not a valid key!" % key, SyntaxWarning,
                         stacklevel)
                full = m.group()
                s = s.replace(full, '%' + full)
        if 'KEY' not in checked:
            return safe_modulo(s, meta, checked=checked + 'KEY',
                               print_warning=print_warning,
                               stacklevel=stacklevel)
        if not isinstance(meta, dict) or 'VALUE' in checked:
            raise
        s = re.sub(r"""(?<!%)(%%)*%(?!%) # uneven number of %
                    \s*(\w|$)         # format strings""", '%\g<0>', s,
                   flags=re.VERBOSE)
        return safe_modulo(s, meta, checked=checked + 'VALUE',
                           print_warning=print_warning, stacklevel=stacklevel)


class DocstringProcessor(object):
    """Class that is intended to process docstrings

    It is, but only to minor extends, inspired by the
    :class:`matplotlib.docstring.Substitution` class.

    Examples
    --------
    Create docstring processor via::

        >>> from docrep import DocstringProcessor
        >>> d = DocstringProcessor(doc_key='My doc string')

    And then use it as a decorator to process the docstring::

        >>> @d
        ... def doc_test():
        ...     '''That's %(doc_key)s'''
        ...     pass

        >>> print(doc_test.__doc__)
        That's My doc string

    Use the :meth:`get_sectionsf` method to extract Parameter sections (or
    others) form the docstring for later usage (and make sure, that the
    docstring is dedented)::

        >>> @d.get_sectionsf('docstring_example',
        ...                  sections=['Parameters', 'Examples'])
        ... @d.dedent
        ... def doc_test(a=1, b=2):
        ...     '''
        ...     That's %(doc_key)s
        ...
        ...     Parameters
        ...     ----------
        ...     a: int, optional
        ...         A dummy parameter description
        ...     b: int, optional
        ...         A second dummy parameter
        ...
        ...     Examples
        ...     --------
        ...     Some dummy example doc'''
        ...     print(a)

        >>> @d.dedent
        ... def second_test(a=1, b=2):
        ...     '''
        ...     My second function where I want to use the docstring from
        ...     above
        ...
        ...     Parameters
        ...     ----------
        ...     %(docstring_example.parameters)s
        ...
        ...     Examples
        ...     --------
        ...     %(docstring_example.examples)s'''
        ...     pass

        >>> print(second_test.__doc__)
        My second function where I want to use the docstring from
        above
        <BLANKLINE>
        Parameters
        ----------
        a: int, optional
            A dummy parameter description
        b: int, optional
            A second dummy parameter
        <BLANKLINE>
        Examples
        --------
        Some dummy example doc

    Another example uses non-dedented docstrings::

        >>> @d.get_sectionsf('not_dedented')
        ... def doc_test2(a=1):
        ...     '''That's the summary
        ...
        ...     Parameters
        ...     ----------
        ...     a: int, optional
        ...         A dummy parameter description'''
        ...     print(a)

    These sections must then be used with the :meth:`with_indent` method to
    indent the inserted parameters::

        >>> @d.with_indent(4)
        ... def second_test2(a=1):
        ...     '''
        ...     My second function where I want to use the docstring from
        ...     above
        ...
        ...     Parameters
        ...     ----------
        ...     %(not_dedented.parameters)s'''
        ...     pass


    """

    #: :class:`dict`. Dictionary containing the compiled patterns to identify
    #: the Parameters, Other Parameters, Warnings and Notes sections in a
    #: docstring
    patterns = {}

    #: :class:`dict`. Dictionary containing the parameters that are used in for
    #: substitution.
    params = {}

    #: sections that behave the same as the `Parameter` section by defining a
    #: list
    param_like_sections = ['Parameters', 'Other Parameters', 'Returns',
                           'Raises']
    #: sections that include (possibly not list-like) text
    text_sections = ['Warnings', 'Notes', 'Examples', 'See Also',
                     'References']

    #: The action on how to react on classes in python 2
    #:
    #: When calling::
    #:
    #:     >>> @docstrings
    #:     ... class NewClass(object):
    #:     ...     """%(replacement)s"""
    #:
    #: This normaly raises an AttributeError, because the ``__doc__`` attribute
    #: of a class in python 2 is not writable. This attribute may be one of
    #: ``'ignore', 'raise' or 'warn'``
    python2_classes = 'ignore'

    def __init__(self, *args, **kwargs):
        """
    Parameters
    ----------
    ``*args`` and ``**kwargs``
        Parameters that shall be used for the substitution. Note that you can
        only provide either ``*args`` or ``**kwargs``, furthermore most of the
        methods like `get_sectionsf` require ``**kwargs`` to be provided."""
        if len(args) and len(kwargs):
            raise ValueError("Only positional or keyword args are allowed")
        self.params = args or kwargs
        patterns = {}
        all_sections = self.param_like_sections + self.text_sections
        for section in self.param_like_sections:
            patterns[section] = re.compile(
                '(?s)(?<=%s\n%s\n)(.+?)(?=\n\n\S+|$)' % (
                    section, '-'*len(section)))
        all_sections_patt = '|'.join(
            '%s\n%s\n' % (s, '-'*len(s)) for s in all_sections)
        # examples and see also
        for section in self.text_sections:
            patterns[section] = re.compile(
                '(?s)(?<=%s\n%s\n)(.+?)(?=%s|$)' % (
                    section, '-'*len(section), all_sections_patt))
        self._extended_summary_patt = re.compile(
            '(?s)(.+?)(?=%s|$)' % all_sections_patt)
        self._all_sections_patt = re.compile(all_sections_patt)
        self.patterns = patterns

    def __call__(self, func):
        """
        Substitute in a docstring of a function with :attr:`params`

        Parameters
        ----------
        func: function
            function with the documentation whose sections
            shall be inserted from the :attr:`params` attribute

        See Also
        --------
        dedent: also dedents the doc
        with_indent: also indents the doc"""
        doc = func.__doc__ and safe_modulo(func.__doc__, self.params,
                                           stacklevel=3)
        return self._set_object_doc(func, doc)

    def get_sections(self, s, base,
                     sections=['Parameters', 'Other Parameters']):
        """
        Method that extracts the specified sections out of the given string if
        (and only if) the docstring follows the numpy documentation guidelines
        [1]_. Note that the section either must appear in the
        :attr:`param_like_sections` or the :attr:`text_sections` attribute.

        Parameters
        ----------
        s: str
            Docstring to split
        base: str
            base to use in the :attr:`sections` attribute
        sections: list of str
            sections to look for. Each section must be followed by a newline
            character ('\\n') and a bar of '-' (following the numpy (napoleon)
            docstring conventions).

        Returns
        -------
        str
            The replaced string

        References
        ----------
        .. [1] https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt

        See Also
        --------
        delete_params, keep_params, delete_types, keep_types, delete_kwargs:
            For manipulating the docstring sections
        save_docstring:
            for saving an entire docstring
        """
        params = self.params
        # Remove the summary and dedent the rest
        s = self._remove_summary(s)
        for section in sections:
            key = '%s.%s' % (base, section.lower().replace(' ', '_'))
            params[key] = self._get_section(s, section)
        return s

    def _remove_summary(self, s):
        # if the string does not start with one of the sections, we remove the
        # summary
        if not self._all_sections_patt.match(s.lstrip()):
            # remove the summary
            lines = summary_patt.sub('', s, 1).splitlines()
            # look for the first line with content
            first = next((i for i, l in enumerate(lines) if l.strip()), 0)
            # dedent the lines
            s = dedents('\n' + '\n'.join(lines[first:]))
        return s

    def _get_section(self, s, section):
        try:
            return self.patterns[section].search(s).group(0).rstrip()
        except AttributeError:
            return ''

    def get_sectionsf(self, *args, **kwargs):
        """
        Decorator method to extract sections from a function docstring

        Parameters
        ----------
        ``*args`` and ``**kwargs``
            See the :meth:`get_sections` method. Note, that the first argument
            will be the docstring of the specified function

        Returns
        -------
        function
            Wrapper that takes a function as input and registers its sections
            via the :meth:`get_sections` method"""
        def func(f):
            doc = f.__doc__
            self.get_sections(doc or '', *args, **kwargs)
            return f
        return func

    def _set_object_doc(self, obj, doc, stacklevel=3):
        """Convenience method to set the __doc__ attribute of a python object
        """
        if isinstance(obj, types.MethodType) and six.PY2:
            obj = obj.im_func
        try:
            obj.__doc__ = doc
        except AttributeError:  # probably python2 class
            if (self.python2_classes != 'raise' and
                    (inspect.isclass(obj) and six.PY2)):
                if self.python2_classes == 'warn':
                    warn("Cannot modify docstring of classes in python2!",
                         stacklevel=stacklevel)
            else:
                raise
        return obj

    def dedent(self, func):
        """
        Dedent the docstring of a function and substitute with :attr:`params`

        Parameters
        ----------
        func: function
            function with the documentation to dedent and whose sections
            shall be inserted from the :attr:`params` attribute"""
        doc = func.__doc__ and self.dedents(func.__doc__, stacklevel=4)
        return self._set_object_doc(func, doc)

    def dedents(self, s, stacklevel=3):
        """
        Dedent a string and substitute with the :attr:`params` attribute

        Parameters
        ----------
        s: str
            string to dedent and insert the sections of the :attr:`params`
            attribute
        stacklevel: int
            The stacklevel for the warning raised in :func:`safe_module` when
            encountering an invalid key in the string"""
        s = dedents(s)
        return safe_modulo(s, self.params, stacklevel=stacklevel)

    def with_indent(self, indent=0):
        """
        Substitute in the docstring of a function with indented :attr:`params`

        Parameters
        ----------
        indent: int
            The number of spaces that the substitution should be indented

        Returns
        -------
        function
            Wrapper that takes a function as input and substitutes it's
            ``__doc__`` with the indented versions of :attr:`params`

        See Also
        --------
        with_indents, dedent"""
        def replace(func):
            doc = func.__doc__ and self.with_indents(
                func.__doc__, indent=indent, stacklevel=4)
            return self._set_object_doc(func, doc)
        return replace

    def with_indents(self, s, indent=0, stacklevel=3):
        """
        Substitute a string with the indented :attr:`params`

        Parameters
        ----------
        s: str
            The string in which to substitute
        indent: int
            The number of spaces that the substitution should be indented
        stacklevel: int
            The stacklevel for the warning raised in :func:`safe_module` when
            encountering an invalid key in the string

        Returns
        -------
        str
            The substituted string

        See Also
        --------
        with_indent, dedents"""
        # we make a new dictionary with objects that indent the original
        # strings if necessary. Note that the first line is not indented
        d = {key: _StrWithIndentation(val, indent)
             for key, val in six.iteritems(self.params)}
        return safe_modulo(s, d, stacklevel=stacklevel)

    def delete_params(self, base_key, *params):
        """
        Method to delete a parameter from a parameter documentation.

        This method deletes the given `param` from the `base_key` item in the
        :attr:`params` dictionary and creates a new item with the original
        documentation without the description of the param. This method works
        for the ``'Parameters'`` sections.

        Parameters
        ----------
        base_key: str
            key in the :attr:`params` dictionary
        ``*params``
            str. Parameter identifier of which the documentations shall be
            deleted

        See Also
        --------
        delete_types"""
        self.params[
            base_key + '.no_' + '|'.join(params)] = self.delete_params_s(
                self.params[base_key], params)

    @staticmethod
    def delete_params_s(s, params):
        """
        Delete the given parameters from a string

        Same as :meth:`delete_params` but does not use the :attr:`params`
        dictionary

        Parameters
        ----------
        s: str
            The string of the parameters section
        params: list of str
            The names of the parameters to delete

        Returns
        -------
        str
            The modified string `s` without the descriptions of `params`
        """
        patt = '(?s)' + '|'.join(
            '(?<=\n)' + s + '\s*:.+?\n(?=\S+|$)' for s in params)
        return re.sub(patt, '', '\n' + s.strip() + '\n').strip()

    def delete_kwargs(self, base_key, args=None, kwargs=None):
        """
        Deletes the ``*args`` or ``**kwargs`` part from the parameters section

        Either `args` or `kwargs` must not be None. The resulting key will be
        stored in

        ``base_key + 'no_args'``
            if `args` is not None and `kwargs` is None
        ``base_key + 'no_kwargs'``
            if `args` is None and `kwargs` is not None
        ``base_key + 'no_args_kwargs'``
            if `args` is not None and `kwargs` is not None

        Parameters
        ----------
        base_key: str
            The key in the :attr:`params` attribute to use
        args: None or str
            The string for the args to delete
        kwargs: None or str
            The string for the kwargs to delete

        Notes
        -----
        The type name of `args` in the base has to be like ````*<args>````
        (i.e. the `args` argument preceeded by a ``'*'`` and enclosed by double
        ``'`'``). Similarily, the type name of `kwargs` in `s` has to be like
        ````**<kwargs>````"""
        if not args and not kwargs:
            warn("Neither args nor kwargs are given. I do nothing for %s" % (
                base_key))
            return
        ext = '.no' + ('_args' if args else '') + ('_kwargs' if kwargs else '')
        self.params[base_key + ext] = self.delete_kwargs_s(
            self.params[base_key], args, kwargs)

    @classmethod
    def delete_kwargs_s(cls, s, args=None, kwargs=None):
        """
        Deletes the ``*args`` or ``**kwargs`` part from the parameters section

        Either `args` or `kwargs` must not be None.

        Parameters
        ----------
        s: str
            The string to delete the args and kwargs from
        args: None or str
            The string for the args to delete
        kwargs: None or str
            The string for the kwargs to delete

        Notes
        -----
        The type name of `args` in `s` has to be like ````*<args>```` (i.e. the
        `args` argument preceeded by a ``'*'`` and enclosed by double ``'`'``).
        Similarily, the type name of `kwargs` in `s` has to be like
        ````**<kwargs>````"""
        if not args and not kwargs:
            return s
        types = []
        if args is not None:
            types.append('`?`?\*%s`?`?' % args)
        if kwargs is not None:
            types.append('`?`?\*\*%s`?`?' % kwargs)
        return cls.delete_types_s(s, types)

    def delete_types(self, base_key, out_key, *types):
        """
        Method to delete a parameter from a parameter documentation.

        This method deletes the given `param` from the `base_key` item in the
        :attr:`params` dictionary and creates a new item with the original
        documentation without the description of the param. This method works
        for ``'Results'`` like sections.

        Parameters
        ----------
        base_key: str
            key in the :attr:`params` dictionary
        out_key: str
            Extension for the base key (the final key will be like
            ``'%s.%s' % (base_key, out_key)``
        ``*types``
            str. The type identifier of which the documentations shall deleted

        See Also
        --------
        delete_params"""
        self.params['%s.%s' % (base_key, out_key)] = self.delete_types_s(
            self.params[base_key], types)

    @staticmethod
    def delete_types_s(s, types):
        """
        Delete the given types from a string

        Same as :meth:`delete_types` but does not use the :attr:`params`
        dictionary

        Parameters
        ----------
        s: str
            The string of the returns like section
        types: list of str
            The type identifiers to delete

        Returns
        -------
        str
            The modified string `s` without the descriptions of `types`
        """
        patt = '(?s)' + '|'.join(
            '(?<=\n)' + s + '\n.+?\n(?=\S+|$)' for s in types)
        return re.sub(patt, '', '\n' + s.strip() + '\n',).strip()

    def keep_params(self, base_key, *params):
        """
        Method to keep only specific parameters from a parameter documentation.

        This method extracts the given `param` from the `base_key` item in the
        :attr:`params` dictionary and creates a new item with the original
        documentation with only the description of the param. This method works
        for ``'Parameters'`` like sections.

        Parameters
        ----------
        base_key: str
            key in the :attr:`params` dictionary
        ``*params``
            str. Parameter identifier of which the documentations shall be
            in the new section

        See Also
        --------
        keep_types"""
        self.params[base_key + '.' + '|'.join(params)] = self.keep_params_s(
            self.params[base_key], params)

    @staticmethod
    def keep_params_s(s, params):
        """
        Keep the given parameters from a string

        Same as :meth:`keep_params` but does not use the :attr:`params`
        dictionary

        Parameters
        ----------
        s: str
            The string of the parameters like section
        params: list of str
            The parameter names to keep

        Returns
        -------
        str
            The modified string `s` with only the descriptions of `params`
        """
        patt = '(?s)' + '|'.join(
            '(?<=\n)' + s + '\s*:.+?\n(?=\S+|$)' for s in params)
        return ''.join(re.findall(patt, '\n' + s.strip() + '\n')).rstrip()

    def keep_types(self, base_key, out_key, *types):
        """
        Method to keep only specific parameters from a parameter documentation.

        This method extracts the given `type` from the `base_key` item in the
        :attr:`params` dictionary and creates a new item with the original
        documentation with only the description of the type. This method works
        for the ``'Results'`` sections.

        Parameters
        ----------
        base_key: str
            key in the :attr:`params` dictionary
        out_key: str
            Extension for the base key (the final key will be like
            ``'%s.%s' % (base_key, out_key)``
        ``*types``
            str. The type identifier of which the documentations shall be
            in the new section

        See Also
        --------
        keep_params"""
        self.params['%s.%s' % (base_key, out_key)] = self.keep_types_s(
            self.params[base_key], types)

    @staticmethod
    def keep_types_s(s, types):
        """
        Keep the given types from a string

        Same as :meth:`keep_types` but does not use the :attr:`params`
        dictionary

        Parameters
        ----------
        s: str
            The string of the returns like section
        types: list of str
            The type identifiers to keep

        Returns
        -------
        str
            The modified string `s` with only the descriptions of `types`
        """
        patt = '|'.join('(?<=\n)' + s + '\n(?s).+?\n(?=\S+|$)' for s in types)
        return ''.join(re.findall(patt, '\n' + s.strip() + '\n')).rstrip()

    def save_docstring(self, key):
        """
        Descriptor method to save a docstring from a function

        Like the :meth:`get_sectionsf` method this method serves as a
        descriptor for functions but saves the entire docstring"""
        def func(f):
            self.params[key] = f.__doc__ or ''
            return f
        return func

    def get_summary(self, s, base=None):
        """
        Get the summary of the given docstring

        This method extracts the summary from the given docstring `s` which is
        basicly the part until two newlines appear

        Parameters
        ----------
        s: str
            The docstring to use
        base: str or None
            A key under which the summary shall be stored in the :attr:`params`
            attribute. If not None, the summary will be stored in
            ``base + '.summary'``. Otherwise, it will not be stored at all

        Returns
        -------
        str
            The extracted summary"""
        summary = summary_patt.search(s).group()
        if base is not None:
            self.params[base + '.summary'] = summary
        return summary

    def get_summaryf(self, *args, **kwargs):
        """
        Extract the summary from a function docstring

        Parameters
        ----------
        ``*args`` and ``**kwargs``
            See the :meth:`get_summary` method. Note, that the first argument
            will be the docstring of the specified function

        Returns
        -------
        function
            Wrapper that takes a function as input and registers its summary
            via the :meth:`get_summary` method"""
        def func(f):
            doc = f.__doc__
            self.get_summary(doc or '', *args, **kwargs)
            return f
        return func

    def get_extended_summary(self, s, base=None):
        """Get the extended summary from a docstring

        This here is the extended summary

        Parameters
        ----------
        s: str
            The docstring to use
        base: str or None
            A key under which the summary shall be stored in the :attr:`params`
            attribute. If not None, the summary will be stored in
            ``base + '.summary_ext'``. Otherwise, it will not be stored at
            all

        Returns
        -------
        str
            The extracted extended summary"""
        # Remove the summary and dedent
        s = self._remove_summary(s)
        ret = ''
        if not self._all_sections_patt.match(s):
            m = self._extended_summary_patt.match(s)
            if m is not None:
                ret = m.group().strip()
        if base is not None:
            self.params[base + '.summary_ext'] = ret
        return ret

    def get_extended_summaryf(self, *args, **kwargs):
        """Extract the extended summary from a function docstring

        This function can be used as a decorator to extract the extended
        summary of a function docstring (similar to :meth:`get_sectionsf`).

        Parameters
        ----------
        ``*args`` and ``**kwargs``
            See the :meth:`get_extended_summary` method. Note, that the first
            argument will be the docstring of the specified function

        Returns
        -------
        function
            Wrapper that takes a function as input and registers its summary
            via the :meth:`get_extended_summary` method"""
        def func(f):
            doc = f.__doc__
            self.get_extended_summary(doc or '', *args, **kwargs)
            return f
        return func

    def get_full_description(self, s, base=None):
        """Get the full description from a docstring

        This here and the line above is the full description (i.e. the
        combination of the :meth:`get_summary` and the
        :meth:`get_extended_summary`) output

        Parameters
        ----------
        s: str
            The docstring to use
        base: str or None
            A key under which the description shall be stored in the
            :attr:`params` attribute. If not None, the summary will be stored
            in ``base + '.full_desc'``. Otherwise, it will not be stored
            at all

        Returns
        -------
        str
            The extracted full description"""
        summary = self.get_summary(s)
        extended_summary = self.get_extended_summary(s)
        ret = (summary + '\n\n' + extended_summary).strip()
        if base is not None:
            self.params[base + '.full_desc'] = ret
        return ret

    def get_full_descriptionf(self, *args, **kwargs):
        """Extract the full description from a function docstring

        This function can be used as a decorator to extract the full
        descriptions of a function docstring (similar to
        :meth:`get_sectionsf`).

        Parameters
        ----------
        ``*args`` and ``**kwargs``
            See the :meth:`get_full_description` method. Note, that the first
            argument will be the docstring of the specified function

        Returns
        -------
        function
            Wrapper that takes a function as input and registers its summary
            via the :meth:`get_full_description` method"""
        def func(f):
            doc = f.__doc__
            self.get_full_description(doc or '', *args, **kwargs)
            return f
        return func



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_array_equal, assert_allclose
import kdcount.correlate as correlate
import os
import pytest

setup_logging()

def make_corrfunc_input(data, cosmo):
    ra = gather_data(data, 'RA')
    dec = gather_data(data, 'DEC')
    z = gather_data(data, 'Redshift')
    rdist = cosmo.comoving_distance(z)
    return numpy.vstack([ra, dec, rdist]).T

def gather_data(source, name):
    return numpy.concatenate(source.comm.allgather(source[name].compute()), axis=0)

def generate_sim_data(seed, comm):
    return UniformCatalog(nbar=3e-4, BoxSize=512., seed=seed, comm=comm)

def generate_survey_data(seed, comm):

    # make the data
    cosmo = cosmology.Planck15
    d = UniformCatalog(nbar=3e-4, BoxSize=512., seed=seed, comm=comm)
    d['RA'], d['DEC'], d['Redshift'] = transform.CartesianToSky(d['Position'], cosmo)

    # make the randoms (ensure nbar is high enough to not have missing values)
    r = UniformCatalog(nbar=3e-4, BoxSize=512., seed=seed*2, comm=comm)
    r['RA'], r['DEC'], r['Redshift'] = transform.CartesianToSky(r['Position'], cosmo)

    return d, r

def reference_survey_tpcf(data1, randoms1, edges, pimax, data2=None, randoms2=None):
    """Reference projected 2PCF using Corrfunc"""
    from Corrfunc.utils import convert_3d_counts_to_cf, convert_rp_pi_counts_to_wp
    from Corrfunc.mocks import DDrppi_mocks

    if data2 is None:
        data2 = data1
    if randoms2 is None:
        randoms2 = randoms1

    # updack the columns to pass to Corrfunc
    ra_d1, dec_d1, r_d1 = data1.T
    ra_r1, dec_r1, r_r1 = randoms1.T
    ra_d2, dec_d2, r_d2 = data2.T
    ra_r2, dec_r2, r_r2 = randoms2.T

    # the Corrfunc keywords
    kws = {}
    kws['cosmology'] = 1
    kws['nthreads'] = 1
    kws['pimax'] = pimax
    kws['binfile'] = edges
    kws['is_comoving_dist'] = True

    # do the pair counting
    D1D2 = DDrppi_mocks(0, RA1=ra_d1, DEC1=dec_d1, CZ1=r_d1, RA2=ra_d2, DEC2=dec_d2, CZ2=r_d2, **kws)
    D1R2 = DDrppi_mocks(0, RA1=ra_d1, DEC1=dec_d1, CZ1=r_d1, RA2=ra_r2, DEC2=dec_r2, CZ2=r_r2, **kws)
    D2R1 = DDrppi_mocks(0, RA1=ra_d2, DEC1=dec_d2, CZ1=r_d2, RA2=ra_r1, DEC2=dec_r1, CZ2=r_r1, **kws)
    R1R2 = DDrppi_mocks(0, RA1=ra_r1, DEC1=dec_r1, CZ1=r_r1, RA2=ra_r2, DEC2=dec_r2, CZ2=r_r2, **kws)

    # combine using Landy-Szalay
    ND1 = len(ra_d1); ND2 = len(ra_d2)
    NR1 = len(ra_r1); NR2 = len(ra_r2)
    CF = convert_3d_counts_to_cf(ND1, ND2, NR1, NR2, D1D2, D1R2, D2R1, R1R2).reshape((-1,pimax))
    wp = convert_rp_pi_counts_to_wp(ND1, ND2, NR1, NR2, D1D2, D1R2, D2R1, R1R2, len(edges)-1, pimax)

    D1D2 = D1D2.reshape((-1,pimax))
    D1R2 = D1R2.reshape((-1,pimax))
    D2R1 = D2R1.reshape((-1,pimax))
    R1R2 = R1R2.reshape((-1,pimax))
    return D1D2, D1R2, D2R1, R1R2, CF, wp

def reference_sim_tpcf(pos1, redges, pimax, BoxSize, randoms=None, pos2=None):
    """Refernce projected 2PCF using halotools"""
    from halotools.mock_observables import rp_pi_tpcf

    pi_bins = numpy.linspace(0, pimax, int(pimax)+1)
    estimator = 'Natural' if randoms is None else 'Landy-Szalay'
    do_auto = True if pos2 is None else False
    return rp_pi_tpcf(pos1, redges, pi_bins, period=BoxSize, sample2=pos2,
                    randoms=randoms, estimator=estimator, do_auto=do_auto)

@MPITest([4])
def test_sim_periodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 20, 10)

    # compute 2PCF
    pimax = 50.
    r = SimulationBox2PCF('projected', source, redges, pimax=pimax, periodic=True)

    # verify with halotools
    pos = gather_data(source, "Position")
    cf = reference_sim_tpcf(pos, redges, pimax, source.attrs['BoxSize'])
    assert_allclose(cf, r.corr['corr'])

@MPITest([4])
def test_sim_nonperiodic_auto(comm):

    # uniform source of particles
    source = generate_sim_data(seed=42, comm=comm)
    randoms = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 20, 10)

    # compute 2PCF
    pimax = 50.
    r = SimulationBox2PCF('projected', source, redges, periodic=False, pimax=pimax, randoms1=randoms)

    # verify with halotools
    pos_d = gather_data(source, "Position")
    pos_r = gather_data(randoms, "Position")
    cf = reference_sim_tpcf(pos_d, redges, pimax, None, randoms=pos_r)
    assert_allclose(cf, r.corr['corr'], rtol=1e-5, atol=1e-3)

@MPITest([4])
def test_sim_periodic_cross(comm):

    # uniform source of particles
    data1 = generate_sim_data(seed=42, comm=comm)
    data2 = generate_sim_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 20, 10)

    # compute 2PCF
    pimax = 50.
    r = SimulationBox2PCF('projected', data1, redges, pimax=pimax, periodic=True, data2=data2)

    # verify with halotools
    pos1 = gather_data(data1, "Position")
    pos2 = gather_data(data2, "Position")
    cf = reference_sim_tpcf(pos1, redges, pimax, data1.attrs['BoxSize'], pos2=pos2)
    assert_allclose(cf, r.corr['corr'])
    r.save('paircount-test.json')
    r2 = SimulationBox2PCF.load('paircount-test.json', comm=comm)

    assert_array_equal(r.corr.data, r2.corr.data)
    assert_array_equal(r.wp.data, r2.wp.data)
    if comm.rank == 0: os.remove('paircount-test.json')


@MPITest([4])
def test_survey_cross(comm):
    cosmo = cosmology.Planck15

    # data and randoms
    data1, randoms1 = generate_survey_data(seed=42, comm=comm)
    data2, randoms2 = generate_survey_data(seed=84, comm=comm)

    # make the bin edges
    redges = numpy.linspace(0.01, 5.0, 2)
    pimax = 10

    # compute 2PCF
    r = SurveyData2PCF('projected', data1, randoms1, redges, pimax=pimax, cosmo=cosmo, data2=data2, randoms2=randoms2)

    # run Corrfunc to verify
    data1 = make_corrfunc_input(data1, cosmo)
    randoms1 = make_corrfunc_input(randoms1, cosmo)
    data2 = make_corrfunc_input(data2, cosmo)
    randoms2 = make_corrfunc_input(randoms2, cosmo)
    D1D2, D1R2, D2R1, R1R2, cf, wp = reference_survey_tpcf(data1, randoms1, redges, pimax,
                                                        data2=data2, randoms2=randoms2)

    # verify pair counts and CF
    assert_allclose(cf, r.corr['corr'])
    assert_allclose(wp, r.wp['corr'])
    assert_allclose(D1D2['npairs'], r.D1D2['npairs'])
    assert_allclose(D1R2['npairs'], r.D1R2['npairs'])
    assert_allclose(D2R1['npairs'], r.D2R1['npairs'])
    assert_allclose(R1R2['npairs'], r.R1R2['npairs'])

    r.save('paircount-test.json')
    r2 = SurveyData2PCF.load('paircount-test.json', comm=comm)
    assert_array_equal(r.corr.data, r2.corr.data)
    assert_array_equal(r.wp.data, r2.wp.data)
    if comm.rank == 0: os.remove('paircount-test.json')



from nbodykit.base.catalog import CatalogSource
from nbodykit.utils import is_structured_array
from nbodykit import CurrentMPIComm
from astropy.table import Table
import numpy

class ArrayCatalog(CatalogSource):
    """
    A CatalogSource initialized from an in-memory :obj:`dict`,
    structured :class:`numpy.ndarray`, or :class:`astropy.table.Table`.

    Parameters
    ----------
    data : obj:`dict`, :class:`numpy.ndarray`, :class:`astropy.table.Table`
        a dictionary, structured ndarray, or astropy Table; items are
        interpreted as the columns of the catalog; the length of any item is
        used as the size of the catalog.
    comm : MPI Communicator, optional
        the MPI communicator instance; default (``None``) sets to the
        current communicator
    **kwargs :
        additional keywords to store as meta-data in :attr:`attrs`
    """
    @CurrentMPIComm.enable
    def __init__(self, data, comm=None, **kwargs):

        # convert astropy Tables to structured numpy arrays
        if isinstance(data, Table):
            data = data.as_array()

        # check for structured data
        if not isinstance(data, dict):
            if not is_structured_array(data):
                raise ValueError(("input data to ArrayCatalog must have a "
                                   "structured data type with fields"))

        self.comm    = comm
        self._source = data

        # compute the data type
        if hasattr(data, 'dtype'):
            keys = sorted(data.dtype.names)
        else:
            keys = sorted(data.keys())
        dtype = numpy.dtype([(key, (data[key].dtype, data[key].shape[1:])) for key in keys])
        self._dtype = dtype

        # verify data types are the same
        dtypes = self.comm.gather(dtype, root=0)
        if self.comm.rank == 0:
            if any(dt != dtypes[0] for dt in dtypes):
                raise ValueError("mismatch between dtypes across ranks in Array")

        # the local size
        self._size = len(self._source[keys[0]])

        for key in keys:
            if len(self._source[key]) != self._size:
                raise ValueError("column `%s` and column `%s` has different size" % (keys[0], key))

        # update the meta-data
        self.attrs.update(kwargs)

        CatalogSource.__init__(self, comm=comm)

    @property
    def hardcolumns(self):
        """
        The union of the columns in the file and any transformed columns.
        """
        defaults = CatalogSource.hardcolumns.fget(self)
        return list(self._dtype.names) + defaults

    def get_hardcolumn(self, col):
        """
        Return a column from the underlying data array/dict.

        Columns are returned as dask arrays.
        """
        if col in self._dtype.names:
            return self.make_column(self._source[col])
        else:
            return CatalogSource.get_hardcolumn(self, col)



# catalogs for different file types
from .file import CSVCatalog
from .file import BinaryCatalog
from .file import BigFileCatalog
from .file import HDFCatalog
from .file import TPMBinaryCatalog
from .file import FITSCatalog
from .file import Gadget1Catalog

from .array import ArrayCatalog
from .lognormal import LogNormalCatalog
from .uniform import UniformCatalog, RandomCatalog
from .halos import HaloCatalog
from .species import MultipleSpeciesCatalog
from .decomposed import DecomposedCatalog

__all__ = ['CSVCatalog',
           'BinaryCatalog',
           'BigFileCatalog',
           'HDFCatalog',
           'TPMBinaryCatalog',
           'FITSCatalog',
           'Gadget1Catalog',
           'ArrayCatalog',
           'LogNormalCatalog',
           'UniformCatalog', 'RandomCatalog',
           'HaloCatalog',
           'DecomposedCatalog',
           'MultipleSpeciesCatalog']



from .base import PairCountBase, verify_input_sources
import numpy
import logging
from six import string_types

class SimulationBoxPairCount(PairCountBase):
    r"""
    Count (weighted) pairs of objects in a simulation box
    as a function of :math:`r`, :math:`(r,\mu)`, :math:`(r_p, \pi)`, or
    :math:`\theta` using the :mod:`Corrfunc` package.

    See the Notes below for the allowed coordinate dimensions.

    The default weighting scheme uses the product of the weights for each
    object in a pair.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~SimulationBoxPairCount.run` for the attributes storing the
    results.

    .. note::

        The algorithm expects the positions of particles in a simulation box to
        be the Cartesian ``x``, ``y``, and ``z`` vectors. To compute
        pair counts on survey data, using right ascension, declination, and
        redshift, see :class:`~nbodykit.algorithms.SurveyDataPairCount`.

    Parameters
    ----------
    mode : '1d', '2d', 'projected', 'angular'
        compute pair counts as a function of the specified coordinate basis;
        see the Notes section below for specifics
    first : CatalogSource
        the first source of particles, providing the 'Position' column
    edges : array_like
        the separation bin edges along the first coordinate dimension;
        depending on ``mode``, the options are :math:`r`, :math:`r_p`, or
        :math:`\theta`. Expected units for distances are :math:`\mathrm{Mpc}/h`
        and degrees for angles. Length of nbins+1
    BoxSize : float, 3-vector, optional
        the size of the box; if 'BoxSize' is not provided in the source
        'attrs', it must be provided here
    periodic : bool, optional
        whether to use periodic boundary conditions
    second : CatalogSource, optional
        the second source of particles to cross-correlate
    los : {'x', 'y', 'z'}, int, optional
        the axis of the simulation box to treat as the line-of-sight direction;
        this can be provided as string identifying one of 'x', 'y', 'z' or
        the equivalent integer number of the axis
    Nmu : int, optional
        the number of :math:`\mu` bins, ranging from 0 to 1; requred if
        ``mode='2d'``
    pimax : float, optional
        The maximum separation along the line-of-sight when ``mode='projected'``.
        Distances along the :math:`\pi` direction are binned with unit
        depth. For instance, if ``pimax=40``, then 40 bins will be created
        along the :math:`\pi` direction.
    weight : str, optional
        the name of the column in the source specifying the particle weights
    show_progress : bool, optional
        if ``True``, perform the pair counting calculation in 10 iterations,
        logging the progress after each iteration; this is useful for
        understanding the scaling of the code
    **config : key/value pairs
        additional keywords to pass to the :mod:`Corrfunc` function

    Notes
    -----
    This class can compute pair counts using several different coordinate
    choices, based on the value of the input argument ``mode``. The choices
    are:

    * ``mode='1d'`` : compute pairs as a function of the 3D separation :math:`r`
    * ``mode='2d'`` : compute pairs as a function of the 3D separation :math:`r`
      and the cosine of the angle to the line-of-sight, :math:`\mu`
    * ``mode='projected'`` : compute pairs as a function of distance perpendicular
      and parallel to the line-of-sight, :math:`r_p` and :math:`\pi`
    * ``mode='angular'`` : compute pairs as a function of angle on the sky, :math:`\theta`

    For angular pair counts, the observer is placed at the center of the
    box when converting Cartesian coordinates to angular coordinates on the
    unit sphere.
    """
    logger = logging.getLogger('SimulationBoxPairCount')

    def __init__(self, mode, first, edges, BoxSize=None, periodic=True,
                    second=None, los='z', Nmu=None, pimax=None,
                    weight='Weight', show_progress=False, **config):

        # check input 'los'
        if isinstance(los, string_types):
            if not los in 'xyz':
                raise ValueError("``los`` should be one of 'x', 'y', 'z'")
            los = 'xyz'.index(los)

        if isinstance(los, int):
            if los < 0: los += 3

        if los not in [0,1,2]:
            raise ValueError("``los`` should be either ['x', 'y', 'z'] or [0,1,2]")

        # verify the input sources
        required_cols = ['Position', weight]
        BoxSize = verify_input_sources(first, second, BoxSize, required_cols)

        # init the base class (this verifies input arguments)
        PairCountBase.__init__(self, mode, edges, first, second, Nmu, pimax, weight, show_progress)

        # save the rest of the meta-data
        self.attrs['BoxSize'] = BoxSize
        self.attrs['periodic'] = periodic
        self.attrs['weight'] = weight
        self.attrs['config'] = config
        self.attrs['los'] = los

        # test maximum separation and periodic boundary conditions
        if periodic and mode != 'angular':
            min_box_side = 0.5*self.attrs['BoxSize'].min()
            if numpy.amax(edges) > min_box_side or mode == 'projected' and pimax > min_box_side:
                raise ValueError(("periodic pair counts cannot be computed for Rmax > BoxSize/2"))

        # run the algorithm
        self.run()

    def run(self):
        """
        Calculate the pair counts in a simulation box.
        This adds the following attributes to the class:

        - :attr:`SimulationBoxPairCount.pairs`

        Attributes
        ----------
        pairs : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object holding the pair count results.
            The coordinate grid will be ``(r,)``, ``(r,mu)``, ``(rp, pi)``,
            or ``(theta,)`` when ``mode`` is '1d', '2d', 'projected', 'angular',
            respectively.

            The BinnedStatistic stores the following variables:

            - ``r``, ``rp``, or ``theta`` : the mean separation value in the bin
            - ``npairs``: the number of pairs in the bin
            - ``wnpairs``: the average weight value in the bin; each pair
              contributes the product of the individual weight values
        """
        # setup
        mode = self.attrs['mode']
        BoxSize = self.attrs['BoxSize']
        first, second = self.first, self.second
        attrs = self.attrs.copy()

        # determine the axes order
        # NOTE: LOS axis should be final column
        los = attrs['los']
        axes_order = [i for i in [0,1,2] if i != los] + [los]

        # compute the max cartesian distance for smoothing
        smoothing = numpy.max(attrs['edges'])
        if mode == 'projected':
            smoothing = numpy.sqrt(smoothing**2 + attrs['pimax']**2)
        elif mode == 'angular':
            smoothing = 2 * numpy.sin(0.5 * numpy.deg2rad(smoothing))

        # if not angular, decompose sim box data (x,y,z)
        if mode != 'angular':
            from .domain import decompose_box_data

            # domain decompose the data
            (pos1, w1), (pos2, w2) = decompose_box_data(first, second, attrs,
                                                        self.logger, smoothing)

            # reorder to make LOS last column
            pos1 = pos1[:,axes_order]
            pos2 = pos2[:,axes_order]

        # NOTE: if doing angular, shift observer to box center and use RA, DEC
        # go from (x,y,z) to (ra,dec), using observer in the middle of the box
        else:
            from nbodykit.transform import CartesianToEquatorial
            from nbodykit.utils import get_data_bounds
            from .domain import decompose_survey_data

            # make sure Position is shifted to an observer at box center
            # and then convert to RA,DEC
            pos1 = shift_to_box_center(first['Position'], BoxSize, self.comm)
            first['ra'], first['dec'] = CartesianToEquatorial(pos1)

            # do the same thing for second source
            if second is not None and second is not first:
                pos2 = shift_to_box_center(second['Position'], BoxSize, self.comm)
                second['ra'], second['dec'] = CartesianToEquatorial(pos2)

            # domain decompose the data
            attrs['ra'], attrs['dec'] = 'ra', 'dec'
            (pos1, w1), (pos2, w2) = decompose_survey_data(first, second, attrs,
                                                            self.logger, smoothing, angular=True)

        # get the Corrfunc callable based on mode
        kws = {k:attrs[k] for k in ['periodic', 'BoxSize', 'show_progress']}
        kws['comm'] = self.comm

        if attrs['mode'] == '1d':
            from .corrfunc.theory import DD
            func = DD(attrs['edges'], **kws)

        elif attrs['mode'] == '2d':
            from .corrfunc.theory import DDsmu
            func = DDsmu(attrs['edges'], attrs['Nmu'], **kws)

        elif attrs['mode'] == 'projected':
            from .corrfunc.theory import DDrppi
            func = DDrppi(attrs['edges'], attrs['pimax'], **kws)

        elif attrs['mode'] == 'angular':
            from .corrfunc.mocks import DDtheta_mocks
            func = DDtheta_mocks(attrs['edges'], comm=self.comm, show_progress=attrs['show_progress'])

        # do the calculation
        self.pairs = func(pos1, w1, pos2, w2, **attrs['config'])

        self.pairs.attrs['total_wnpairs'] = self.attrs['total_wnpairs']

def shift_to_box_center(pos, BoxSize, comm):
    """
    Find the bounds of the input position array, and if needed, shift the
    position to an observer at the box center.

    Position should be bounded by [0, BoxSize] or [-BoxSize/2, BoxSize/2];
    if not, an exception will be raised.

    Parameters
    ----------
    pos : dask array
        the dask array holding the Position
    BoxSize : array_like
        the size of the box
    comm :
        the MPI communicator

    Returns
    -------
    pos : dask array
        the position array, shifted such that observer is in the box center
    """
    from nbodykit.utils import get_data_bounds

    # make BoxSize is a 3-vector
    _BoxSize = numpy.empty(3)
    _BoxSize[:] = BoxSize

    # get min/max of position (3-vectors)
    pos_min, pos_max = get_data_bounds(pos, comm)

    # Position is [0, BoxSize] --> shift to center of box
    if (pos_min > 0.).all() and (pos_max < _BoxSize).all():
        pos -= 0.5 * _BoxSize
    elif (pos_min > -0.5*_BoxSize).all() and (pos_max < 0.5*_BoxSize).all():
        pass
    else:
        raise ValueError("input Position should be bounded by [0,BoxSize] or [-BoxSize/2,BoxSize/2]")

    return pos



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

import pytest

# debug logging
setup_logging("debug")


@MPITest([1])
def test_missing_ranks(comm):

    with CurrentMPIComm.enter(comm):
        cpus_per_task = 2
        with pytest.raises(ValueError):
            with TaskManager(cpus_per_task, debug=True, use_all_cpus=True) as tm:
                pass

@MPITest([2])
def test_no_workers(comm):

    with CurrentMPIComm.enter(comm):
        cpus_per_task = 2
        with pytest.raises(ValueError):
            with TaskManager(cpus_per_task, debug=True, use_all_cpus=False) as tm:
                pass


@MPITest([2, 4])
def test_iterate(comm):

    cosmo = cosmology.Planck15

    cpus_per_task = 2
    with CurrentMPIComm.enter(comm):
        with TaskManager(cpus_per_task, debug=True, use_all_cpus=True) as tm:

            try:
                for seed in tm.iterate([0, 1, 2]):

                    # uniform particles
                    source = UniformCatalog(nbar=3e-7, BoxSize=1380., seed=seed)

                    # compute P(k,mu) and multipoles
                    r = FFTPower(source, mode='2d', Nmesh=8, poles=[0,2,4])

                    # and save
                    output = "./test_batch_uniform_seed%d.json" % seed
                    r.save(output)

            except Exception as e:
                print(e)
                raise

@MPITest([4])
def test_map(comm):

    cosmo = cosmology.Planck15

    cpus_per_task = 2

    def fftpower(seed):

        # uniform particles shall not use comm, since this is testing
        # TaskManager setting CurrentMPIComm
        source = UniformCatalog(nbar=3e-7, BoxSize=1380., seed=seed)

        # compute P(k,mu) and multipoles
        r = FFTPower(source, mode='2d', Nmesh=8, poles=[0,2,4])

        # and save
        output = "./test_batch_uniform_seed%d.json" % seed
        r.save(output)

        return seed

    with TaskManager(cpus_per_task, debug=True, use_all_cpus=False, comm=comm) as tm:

        try:
            seeds = [0, 1, 2]
            results = tm.map(fftpower, seeds)
            assert all(seeds[i] == results[i] for i in range(len(seeds)))
        except Exception as e:
            print(e)
            raise



from nbodykit.cosmology.linearnbody import LinearNbody
from numpy.testing import assert_allclose

def test_linear_nbody():
    from nbodykit.cosmology import Cosmology
    cosmo = Cosmology(Omega0_cdm = 0.26, Omega0_b = 0.04,
                      m_ncdm=[0.6], P_k_max=1e-2)

    linearnbody = LinearNbody(cosmo, c_b = 0, c_ncdm_1ev_z0=0)

    a0 = 0.01

    k, q0, p0 = linearnbody.seed_from_synchronous(cosmo, a0)

    k, qt, pt = linearnbody.seed_from_synchronous(cosmo, 1.0)

    a, q, p = linearnbody.integrate(k, q0, p0, [a0, 1.0])

    # shall be within 5% to class solution
    assert_allclose(q[-1] / qt, 1., rtol=0.1)

    # need very high precision for reversibility.
    a, q, p = linearnbody.integrate(k, qt, pt, [1.0, a0], rtol=1e-9)
    a, q, p = linearnbody.integrate(k, q[-1], p[-1], [a0, 1.0], rtol=1e-9)

    assert_allclose(q[-1], qt, rtol=1e-4)
    assert_allclose(p[-1], pt, rtol=1e-4)




import numpy
import logging
from mpi4py import MPI

from nbodykit import CurrentMPIComm
from nbodykit.transform import ConstantArray
from scipy.interpolate import InterpolatedUnivariateSpline

class RedshiftHistogram(object):
    """
    Compute the mean number density as a function of redshift
    :math:`n(z)` from an input CatalogSource of particles.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~RedshiftHistogram.run` for the attributes storing the results.

    .. note::
        The units of the number density are :math:`(\mathrm{Mpc}/h)^{-3}`

    Parameters
    ----------
    source : CatalogSource
        the source of particles holding the redshift column to histogram
    fsky : float
        the sky area fraction, which is used in the volume calculation when
        normalizing :math:`n(z)`
    cosmo : :class:`nbodykit.cosmology.core.Cosmology`
        the cosmological parameters, which are used to compute the volume
        from redshift shells when normalizing :math:`n(z)`
    bins : int or sequence of scalars, optional
        If `bins` is an int, it defines the number of equal-width
        bins in the given range. If `bins` is a sequence, it defines the bin
        edges, including the rightmost edge, allowing for non-uniform bin widths.
        If not provided, Scott's rule is used to estimate the optimal bin width
        from the input data (default)
    redshift : str, optional
        the name of the column specifying the redshift data
    weight : str, optional
        the name of the column specifying weights to use when histogramming the data
    """
    logger = logging.getLogger('RedshiftHistogram')

    def __init__(self, source, fsky, cosmo, bins=None, redshift='Redshift', weight=None):

        # input columns need to be there
        for col in [redshift, weight]:
            if col is not None and col not in source:
                raise ValueError("'%s' column missing from input source in RedshiftHistogram" %col)

        self.comm = source.comm

        # using Scott's rule for binning
        if bins is None:
            h, bins = scotts_bin_width(source.compute(source[redshift]), self.comm)
            if self.comm.rank == 0:
                self.logger.info("using Scott's rule to determine optimal binning; h = %.2e, N_bins = %d" %(h, len(bins)-1))

        # equally spaced bins from min to max val
        elif numpy.isscalar(bins):
            if self.comm.rank == 0:
                self.logger.info("computing %d equally spaced bins" %bins)
            z = source.compute(source[redshift])
            maxval = comm.allreduce(z.max(), op=MPI.MAX)
            minval = comm.allreduce(z.min(), op=MPI.MIN)
            bins = linspace(minval, maxval, bins + 1, endpoint=True)

        self.source = source
        self.cosmo  = cosmo

        self.attrs = {}
        self.attrs['edges'] = bins
        self.attrs['fsky'] = fsky
        self.attrs['redshift'] = redshift
        self.attrs['weight'] = weight
        self.attrs['cosmo'] = dict(cosmo)

        # and run
        self.run()

    def run(self):
        """
        Run the algorithm, which computes the histogram. This function
        does not return anything, but adds the following attributes
        to the class:

        - :attr:`bin_edges`
        - :attr:`bin_centers`
        - :attr:`dV`
        - :attr:`nbar`

        .. note::
            All ranks store the same result attributes.

        Attributes
        ----------
        bin_edges : array_like
            the edges of the redshift bins
        bin_centers : array_like
            the center values of each redshift bin
        dV : array_like
            the volume of each redshift shell in units of :math:`(\mathrm{Mpc}/h)^3`
        nbar : array_like
            the values of the redshift histogram, normalized to
            number density (in units of :math:`(\mathrm{Mpc}/h)^{-3}`)
        """
        edges = self.attrs['edges']

        # get the columns
        redshift = self.source[self.attrs['redshift']]
        if self.attrs['weight'] is not None:
            weight = self.source[self.attrs['weight']]
            if self.comm.rank == 0:
                self.logger.info("computing histogram using weights from '%s' column" %self.attrs['weight'])
        else:
            weight = ConstantArray(1.0, self.source.size)

        # compute the numpy arrays from dask
        redshift, weight = self.source.compute(redshift, weight)

        # do the bin count, using the specified weight values
        dig = numpy.searchsorted(edges, redshift, "right")
        N = numpy.bincount(dig, weights=weight, minlength=len(edges)+1)[1:-1]

        # now sum across all ranks
        N = self.comm.allreduce(N)

        # compute the volume
        if self.comm.rank == 0:
            self.logger.info("using cosmology %s to compute volume in units of (Mpc/h)^3" %str(self.cosmo))
            self.logger.info("sky fraction used in volume calculation: %.4f" %self.attrs['fsky'])
        R_hi = self.cosmo.comoving_distance(edges[1:]) # in Mpc/h
        R_lo = self.cosmo.comoving_distance(edges[:-1]) # in Mpc/h
        dV   = (4./3.)*numpy.pi*(R_hi**3 - R_lo**3) * self.attrs['fsky']

        # store the results
        self.bin_edges   = edges
        self.bin_centers = 0.5*(edges[:-1] + edges[1:])
        self.dV          = dV
        self.nbar        = 1.*N/dV

    def interpolate(self, z):
        """ Interpoalte dndz as a function of redshift. 

            The interpolation acts as a band pass filter, removing small scale
            fluctuations in the estimator.

            Returns
            -------
            n : n(z)
        """
        nofz = InterpolatedUnivariateSpline(self.bin_centers, self.nbar)
        return nofz(z)

    def __getstate__(self):
        state = dict(
                     bin_edges=self.bin_edges,
                     bin_centers=self.bin_centers,
                     dV=self.dV,
                     nbar=self.nbar,
                     attrs=self.attrs)
        return state

    def __setstate__(self, state):
        self.__dict__.update(state)

    def save(self, output):
        """
        Save the RedshiftHistogram result to disk.

        The format is JSON.
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('histogram done; saving result to %s' %output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, output, comm=None):
        """
        Load a saved RedshiftHistogram result.

        The result has been saved to disk with :func:`RedshiftHistogram.save`.
        """
        import json
        from nbodykit.utils import JSONDecoder
        if comm.rank == 0:
            with open(output, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)
        self.__setstate__(state)
        self.comm = comm
        return self

def scotts_bin_width(data, comm):
    r"""
    Return the optimal histogram bin width using Scott's rule,
    defined as:

    .. math::

        h = \sigma \sqrt[3]{\frac{24 * \sqrt{\pi}}{n}}

    .. note::

        This is a collective operation

    Parameters
    ----------
    data : array_like
        the array that we are histograming
    comm :
        the MPI communicator

    Returns
    -------
    dx : float
        the bin spacing
    edges : array_like
        the array holding the bin edges
    """
    # compute the mean
    csum = comm.allreduce(data.sum())
    csize = comm.allreduce(data.size)
    cmean = csum / csize

    # std dev
    rsum = comm.allreduce((abs(data - cmean)**2).sum())
    sigma = (rsum / csize)**0.5

    dx = sigma * (24. * numpy.sqrt(numpy.pi) / csize) ** (1. / 3)
    maxval = comm.allreduce(data.max(), op=MPI.MAX)
    minval = comm.allreduce(data.min(), op=MPI.MIN)

    Nbins = numpy.ceil((maxval - minval) * 1. / dx)
    Nbins = max(1, Nbins)
    edges = minval + dx * numpy.arange(Nbins + 1)
    return dx, edges



from runtests.mpi import MPITest
from nbodykit.io.tpm import TPMBinaryFile
import numpy
import tempfile
import os
import pytest

@MPITest([1])
def test_data(comm):

    tmpfile = tempfile.mktemp()
    with open(tmpfile, 'wb') as ff: 
        
        # generate TPM-format data
        pos = numpy.random.random(size=(1024, 3)).astype('f4')
        vel = numpy.random.random(size=(1024, 3)).astype('f4')
        uid = numpy.arange(1024, dtype='u8')
        hdr = numpy.ones(28, dtype='?')
        
        # write to file
        hdr.tofile(ff)
        pos.tofile(ff); vel.tofile(ff); uid.tofile(ff)
        ff.seek(0)
        
        # read
        f = TPMBinaryFile(ff.name, precision='f4')
        
        # check size
        assert f.size == 1024
        
        # and data
        numpy.testing.assert_almost_equal(pos, f['Position'][:])
        numpy.testing.assert_almost_equal(vel, f['Velocity'][:])
        numpy.testing.assert_almost_equal(uid, f['ID'][:])
        
        # check wrong precision
        with pytest.raises(ValueError): 
            f = TPMBinaryFile(ff.name, precision='f16')
        
    os.remove(tmpfile)


import numpy
import os
from pandas import read_csv
from six import string_types

from .base import FileType
from . import tools

class CSVPartition(object):
    """
    A simple class to convert byte strings of data from a CSV file
    to a pandas DataFrame on demand

    The DataFrame is cached as :attr:`value`, so only a single
    call to :func:`pandas.read_csv` is used
    """
    def __init__(self, filename, offset, blocksize, delimiter, **config):
        """
        Parameters
        ----------
        filename : str
            the file to read data from
        offset : int
            the offset in bytes to start reading at
        blocksize : int
            the size of the bytes block to read
        delimiter : byte str
            how to distinguish separate lines
        **config
            the configuration keywords passed to :func:`pandas.read_csv`
        """
        self.filename  = filename
        self.offset    = offset
        self.blocksize = blocksize
        self.delimiter = delimiter
        self.config    = config

    @property
    def value(self):
        """
        Return the parsed btye string as a DataFrame
        """
        try:
            return self._value
        except AttributeError:
            from io import BytesIO
            from dask.bytes.utils import read_block

            # read the relevant bytes
            with open(self.filename, 'rb') as f:
                block = read_block(f, self.offset, self.blocksize, self.delimiter)

            # parse the byte string
            b = BytesIO()
            b.write(block); b.seek(0)
            self._value = read_csv(b, **self.config)

            return self._value

def make_partitions(filename, blocksize, config, delimiter="\n"):
    """
    Partition a CSV file into blocks, using the preferred blocksize
    in bytes, returning the partititions and number of rows in
    each partition

    This divides the input file into partitions with size
    roughly equal to blocksize, reads the bytes, and counts
    the number of delimiters to compute the size of each block

    Parameters
    ----------
    filename : str
        the name of the CSV file to load
    blocksize : int
        the desired number of bytes per block
    delimiter : str, optional
        the character separating lines; default is
        the newline character
    config : dict
        any keyword options to pass to :func:`pandas.read_csv`

    Returns
    -------
    partitions : list of CSVPartition
        list of objects storing the data content of each file partition,
        stored as a bytestring
    sizes : list of int
        the list of the number of rows in each partition
    """
    from dask.bytes.utils import read_block

    config = config.copy()

    # search for lines separated by this character
    delimiter = delimiter.encode()

    # size in bytes and byte offsets of each partition
    size = os.path.getsize(filename)
    offsets = list(range(0, size, int(blocksize)))

    # skip blank lines
    skip_blank_lines = config.get('skip_blank_lines', True)

    # number of rows to read
    nrows = config.pop('nrows', None)

    sizes = []; partitions = []
    with open(filename, 'rb') as f:
        for i, offset in enumerate(offsets):

            # skiprows only valid for first block
            if i > 0 and 'skiprows' in config:
                config.pop('skiprows')

            # set nrows for this block
            config['nrows'] = nrows

            block = read_block(f, offset, blocksize, delimiter)
            partitions.append(CSVPartition(filename, offset, blocksize, delimiter, **config))

            # count delimiter to get size
            size = block.count(delimiter)

            # account for blank lines
            if skip_blank_lines:
                size -= block.count(delimiter+delimiter)
                if i == 0 and block.startswith(delimiter):
                    size -= 1

            # account for skiprows
            skiprows = config.get('skiprows', 0)
            size -= skiprows

            # account for nrows
            if nrows is not None and nrows > 0:
                if nrows < size:
                    sizes.append(nrows)
                    break
                else:
                    nrows -= size # update for next block

            # manually increase size if at end of the file and no newline
            if i == len(offsets)-1 and not block.endswith(delimiter):
                size += 1

            sizes.append(size)



    return partitions, sizes

def verify_data(path, names, nrows=10, **config):
    """
    Verify the data by reading the first few lines of the specified
    CSV file to determine the data type

    Parameters
    ----------
    path : str
        the name of the CSV file to load
    names : list of str
        the list of the names of the columns in the CSV file
    nrows : int, optional
        the number of rows to read from the file in order
        to infer the data type; default is 10
    **config : key, value pairs
        additional keywords to pass to :func:`pandas.read_csv`

    Returns
    -------
    dtype : dict
        dictionary holding the dtype for each name in `names`
    """
    # read the first few lines to get the the dtype
    try:
        df = read_csv(path, nrows=nrows, names=names, **config)

        if df.isnull().sum().any():
            raise ValueError("'NaN' entries found when reading first %d rows; likely configuration error" %nrows)
        if any(dt == 'O' for dt in df.dtypes):
            raise ValueError("'object' data types found when reading first %d rows; likely configuration error" %nrows)

    except:
        import traceback
        config['names'] = names
        msg = ("error trying to read data with pandas.read_csv; ensure that 'names' matches "
               "the number of columns in the file and the file contains no comments\n")
        msg += "pandas configuration: %s\n" %str(config)
        msg += "\n%s" %(traceback.format_exc())
        raise ValueError(msg)

    toret = {}
    for name in df.columns:
        toret[name] = df[name].dtype
    return toret


class CSVFile(FileType):
    """
    A file object to handle the reading of columns of data from a CSV file.

    Internally, this class partitions the CSV file into chunks, and
    data is only read from the relevant chunks of the file, using
    :func:`pandas.read_csv`.

    This setup provides a significant speed-up when reading
    from the end of the file, since the entirety of the data
    does not need to be read first.

    The class supports any of the configuration keywords that can be
    passed to :func:`pandas.read_csv`

    .. warning::
        This assumes the delimiter for separate lines is the newline
        character and that all columns in the file represent data
        columns (no "index" column when using ``pandas``)

    Parameters
    ----------
    path : str
        the name of the file to load
    names : list of str
        the names of the columns of the csv file; this should give
        names of all the columns in the file -- pass ``usecols``
        to select a subset of columns
    blocksize : int, optional
        the file will be partitioned into blocks of bytes roughly
        of this size
    dtype : dict, str, optional
        if specified as a string, assume all columns have this dtype,
        otherwise; each column can have a dtype entry in the dict;
        if not specified, the data types will be inferred from the file
    usecols : list, optional
        a ``pandas.read_csv``; a subset of ``names`` to store, ignoring
        all other columns
    delim_whitespace : bool, optional
        a ``pandas.read_csv`` keyword; if the CSV file is space-separated,
        set this to ``True``
    **config :
        additional keyword arguments that will be passed to
        :func:`pandas.read_csv`; see the documentation of that
        function for a full list of possible options
    """
    def __init__(self, path, names, blocksize=32*1024*1024, dtype={},
                    usecols=None, delim_whitespace=True, **config):

        self.path      = path
        self.names     = names if usecols is None else usecols
        self.blocksize = blocksize

        # ensure that no index column is passed
        if 'index_col' in config and config['index_col']:
            raise ValueError("'index_col = False' is not supported in CSVFile")
        config['index_col'] = False # no index columns in file

        # manually remove comments
        if 'comment' in config and config['comment'] is not None:
            raise ValueError("please manually remove all comments from file")
        config['comment'] = None # no comments

        # ensure that no header is passed
        if 'header' in config and config['header'] is not None:
            raise ValueError("'header' not equal to None is not supported in CSVFile")
        config['header'] = None # no header

        if isinstance(config.get('skiprows', None), list):
            raise ValueError("only integer values supported for 'skiprows' in CSVFile")

        if 'skipfooter' in config:
            raise ValueError("'skipfooter' not supported in CSVFile")

        # set the read_csv defaults
        if 'sep' in config or 'delimiter' in config:
            delim_whitespace = False
        config['delim_whitespace'] = delim_whitespace
        config['usecols'] = usecols
        config.setdefault('engine', 'c')
        config.setdefault('skip_blank_lines', True)
        self.pandas_config = config.copy()

        # verify the data
        inferred_dtype = verify_data(self.path, names, **self.pandas_config)

        # dtype can also be a string --> apply to all columns
        if isinstance(dtype, string_types):
            dtype = {col:dtype for col in self.names}

        # store the dtype as a list
        self.dtype = []
        for col in self.names:
            if col in dtype:
                dt = dtype[col]
                if not isinstance(dt, numpy.dtype):
                    dt = numpy.dtype(dt)
            elif col in inferred_dtype:
                dt = inferred_dtype[col]
            else:
                raise ValueError("data type for column '%s' cannot be inferred from file" %col)
            self.dtype.append((col, dt))
        self.dtype = numpy.dtype(self.dtype)

        # add the dtype and names to the pandas config
        if config['engine'] == 'c':
            self.pandas_config['dtype'] = {col:self.dtype[col] for col in self.names}
        self.pandas_config['names'] = names

        # make the partitions
        self.partitions, self._sizes = make_partitions(path, blocksize, self.pandas_config)
        self.size = sum(self._sizes)

    def read(self, columns, start, stop, step=1):
        """
        Read the specified column(s) over the given range

        'start' and 'stop' should be between 0 and :attr:`size`,
        which is the total size of the file (in particles)

        Parameters
        ----------
        columns : str, list of str
            the name of the column(s) to return
        start : int
            the row integer to start reading at
        stop : int
            the row integer to stop reading at
        step : int, optional
            the step size to use when reading; default is 1

        Returns
        -------
        numpy.array
            structured array holding the requested columns over
            the specified range of rows
        """
        if isinstance(columns, string_types): columns = [columns]

        toret = []
        for fnum in tools.get_file_slice(self._sizes, start, stop):

            # the local slice
            sl = tools.global_to_local_slice(self._sizes, start, stop, fnum)

            # access the dataframe of this partition
            data = self.partitions[fnum].value

            # slice and convert to a structured array
            data = data[sl[0]:sl[1]]
            data = data[columns]
            toret.append(data.to_records(index=False))

        return numpy.concatenate(toret, axis=0)[::step]



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_array_equal
import pytest

setup_logging()


@MPITest([1, 4])
def test_get_syntax(comm):

    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)
    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    # test either get syntax
    test1 = numpy.random.random(size=len(source1))
    cat['data/test1'] = test1
    assert_array_equal(cat['data/test1'], cat['data']['test1'])

    test2 = numpy.random.random(size=len(source1))
    cat['data']['test2'] = test2
    assert_array_equal(cat['data/test2'], cat['data']['test2'])


@MPITest([1, 4])
def test_columns(comm):

    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)

    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2, BoxSize=512., Nmesh=128)

    assert 'BoxSize' in cat.attrs
    assert 'Nmesh' in cat.attrs

    # check all columns are there
    for col in source1:
        assert 'data/' + col in cat
        assert 'randoms/' + col in cat

@MPITest([1, 4])
def test_bad_input(comm):

    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)

    # need 2 species
    with pytest.raises(ValueError):
        cat = MultipleSpeciesCatalog(['data'], source1)

    # non-unique names
    with pytest.raises(ValueError):
        cat = MultipleSpeciesCatalog(['data', 'data'], source1, source2)

    # mis-matched sizes
    with pytest.raises(ValueError):
        cat = MultipleSpeciesCatalog(['data', 'randoms', 'bad'], source1, source2)

    # bad comm
    with pytest.raises(ValueError):
        source1.comm = None
        cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)


@MPITest([1, 4])
def test_getitem(comm):

    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)

    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    for source, name in zip([source1, source2], ['data', 'randoms']):
        subcat = cat[name] # should be equal to source
        for col in source:
            assert col in subcat
            assert_array_equal(subcat[col].compute(), source[col].compute())
        for k in subcat.attrs:
            assert k in source.attrs


@MPITest([1, 4])
def test_setitem(comm):

    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)

    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    # bad name
    with pytest.raises(ValueError):
        cat['bad'] = numpy.ones(source1.size)

    # bad species name
    with pytest.raises(ValueError):
        cat['bad/test'] = numpy.ones(source1.size)

    test = numpy.ones(source1.size)*10
    cat['data/test'] = test
    assert_array_equal(cat['data/test'].compute(), test)

    # bad size
    with pytest.raises(ValueError):
        cat['data/test'] = test[:10]

@MPITest([1, 4])
def test_bad_slice(comm):

    source1 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=42, comm=comm)
    source2 = UniformCatalog(nbar=3e-5, BoxSize=512., seed=84, comm=comm)

    cat = MultipleSpeciesCatalog(['data', 'randoms'], source1, source2)

    # test with a column slice
    with pytest.raises(ValueError):
        subcat = cat[cat['data/Selection']]

    # test with a slice
    with pytest.raises(ValueError):
        subcat = cat[:10]



from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic
import numpy

class PairCountBase(object):
    """
    An abstract base class for pair counting algorithms.

    Users should use one of the subclasses of this class.
    """
    def __init__(self, mode, edges, first, second, Nmu, pimax, weight, show_progress=False):

        # check input 'mode'
        valid_modes = ['1d', '2d', 'projected', 'angular']
        if mode not in valid_modes:
            raise ValueError("allowed 'mode' values are: %s" % valid_modes)

        # check min edge
        if numpy.min(edges) <= 0.:
            raise ValueError(("the lower edge of the 1st separation bin must "
                              "greater than zero (no self-pairs)"))

        # check mode requirements
        if mode == '2d' and Nmu is None:
            raise ValueError("'Nmu' keyword is required when 'mode' is '2d'")
        if Nmu is not None and mode != '2d':
            raise ValueError("mode should be '2d' if 'Nmu' is specified")
        if mode == 'projected' and pimax is None:
            raise ValueError("'pimax' keyword is required when 'mode' is 'projected'")
        if pimax is not None and mode != 'projected':
            raise ValueError("mode should be 'projected' if 'projected' is specified")
        if mode == 'projected' and pimax < 1.0:
            raise ValueError("'pimax' must be at least 1.0 when 'mode' is 'projected'")

        self.first = first
        self.second = second
        self.comm = first.comm

        # store the meta-data
        self.attrs = {}
        self.attrs['mode'] = mode
        self.attrs['edges'] = edges
        self.attrs['Nmu'] = Nmu
        self.attrs['pimax'] = pimax
        self.attrs['show_progress'] = show_progress

        # store the total size of the sources
        self.attrs['N1'] = first.csize
        self.attrs['N2'] = second.csize if second is not None else None

        if second is None or second is first:
            wpairs1, wpairs2 = self.comm.allreduce(first.compute(first[weight].sum())), self.comm.allreduce(first.compute((first[weight]**2).sum()))
            # for auto excluding self pairs to avoid a biased estimator. The factor 0.5 is by convention.
            # In the end it will cancel out in two point function estimators.
            self.attrs['total_wnpairs'] = 0.5*(wpairs1**2-wpairs2)
            self.attrs['is_cross'] = False
        else:
            wpairs1, wpairs2 = self.comm.allreduce(first.compute(first[weight].sum())), self.comm.allreduce(second.compute(second[weight].sum()))
            self.attrs['total_wnpairs'] = 0.5*wpairs1*wpairs2
            self.attrs['is_cross'] = True

    def __getstate__(self):
        return {'pairs':self.pairs.data, 'attrs':self.attrs}

    def __setstate__(self, state):
        self.__dict__.update(state)
        edges = self.attrs['edges']

        # reconstruct the result based on mode
        kws = {'fields_to_sum' : ['npairs', 'wnpairs']}
        if self.attrs['mode'] == '1d':
            dims, edges = ['r'], [edges]

        elif self.attrs['mode'] == '2d':
            muedges = numpy.linspace(0, 1., self.attrs['Nmu']+1)
            dims, edges = ['r', 'mu'], [edges, muedges]

        elif self.attrs['mode'] == 'projected':
            piedges = numpy.linspace(0, self.attrs['pimax'], self.attrs['pimax']+1)
            dims, edges = ['rp', 'pi'], [edges, piedges]

        elif self.attrs['mode'] == 'angular':
            dims, edges = ['theta'], [edges]

        else:
            valid = ['1d','2d','angular','projected']
            args = (self.attrs['mode'], valid)
            raise ValueError("mode = '%s' should be one of %s" % args)

        # save the result as a BinnedStatistic
        self.pairs = BinnedStatistic(dims, edges, self.pairs, **kws)

    def save(self, output):
        """
        Save result as a JSON file with name ``output``
        """
        import json
        from nbodykit.utils import JSONEncoder

        # only the master rank writes
        if self.comm.rank == 0:
            self.logger.info('measurement done; saving result to %s' % output)

            with open(output, 'w') as ff:
                json.dump(self.__getstate__(), ff, cls=JSONEncoder)

    @classmethod
    @CurrentMPIComm.enable
    def load(cls, output, comm=None):
        """
        Load a result has been saved to disk with :func:`save`.
        """
        import json
        from nbodykit.utils import JSONDecoder
        if comm.rank == 0:
            with open(output, 'r') as ff:
                state = json.load(ff, cls=JSONDecoder)
        else:
            state = None
        state = comm.bcast(state)
        self = object.__new__(cls)
        self.__setstate__(state)
        self.comm = comm
        return self

def verify_input_sources(first, second, BoxSize, required_columns, inspect_boxsize=True):
    """
    Verify that the input source objects have all of the required columns
    and appropriate box size attributes.
    """
    if second is None: second = first

    # check for comm mismatch
    assert second.comm is first.comm, "communicator mismatch between input sources"

    # check required columns
    for source in [first, second]:
        for col in required_columns:
            if col not in source:
                raise ValueError("the column '%s' is missing from input source; cannot do pair count" %col)

    if inspect_boxsize:
        _BoxSize = numpy.zeros(3)
        BoxSize1 = first.attrs.get('BoxSize', None)
        if BoxSize1 is not None:
            _BoxSize[:] = BoxSize1[:]
        if BoxSize is not None:
            _BoxSize[:] = BoxSize
        if (_BoxSize==0.).all():
            raise ValueError("BoxSize must be supplied in the source ``attrs`` or via the ``BoxSize`` keyword")

        BoxSize2 = second.attrs.get('BoxSize', None)
        if BoxSize1 is not None and BoxSize2 is not None:
            if not numpy.array_equal(first.attrs['BoxSize'], BoxSize2):
                raise ValueError("BoxSize mismatch between pair count cross-correlation sources")
            if not numpy.array_equal(first.attrs['BoxSize'], _BoxSize):
                raise ValueError("BoxSize mismatch between sources and the pair count algorithm")

        return _BoxSize



from runtests.mpi import MPITest
from nbodykit.lab import *
from nbodykit import setup_logging

from numpy.testing import assert_allclose

setup_logging()

@MPITest([1,4])
def test_paint(comm):

    cosmo = cosmology.Planck15

    # linear grid
    Plin = cosmology.LinearPower(cosmo, redshift=0.55, transfer='EisensteinHu')
    source = LinearMesh(Plin, Nmesh=64, BoxSize=512, seed=42, comm=comm)

    # compute P(k) from linear grid
    r = FFTPower(source, mode='1d', Nmesh=64, dk=0.01, kmin=0.005)

    # run and get the result
    valid = r.power['modes'] > 0

    # variance of each point is 2*P^2/N_modes
    theory = Plin(r.power['k'][valid])
    errs = (2*theory**2/r.power['modes'][valid])**0.5

    # compute reduced chi-squared of measurement to theory
    chisq = ((r.power['power'][valid].real - theory)/errs)**2
    N = valid.sum()
    red_chisq = chisq.sum() / (N-1)

    # make sure it is less than 1.5 (should be ~1)
    assert red_chisq < 1.5, "reduced chi sq of linear grid measurement = %.3f" %red_chisq



from nbodykit.base.mesh import MeshSource

from nbodykit import CurrentMPIComm
from nbodykit.utils import attrs_to_dict
from pmesh.pm import RealField, ComplexField
import numpy

class ArrayMesh(MeshSource):
    """
    A MeshSource initalized from an in-memory numpy array.

    .. note::
        The in-memory array must be fully hosted by the root rank.

    Parameters
    ----------
    array : numpy.ndarray
        the numpy array holding the field data; this must be fully
        hosted by the rank specified by ``root``
    BoxSize : float, 3-vector
        the size of the box
    root : int, optional
        the root rank holding the array data
    **kwargs :
        additional meta-data to store
    """
    def __repr__(self):
        return "ArrayMesh()"

    @CurrentMPIComm.enable
    def __init__(self, array, BoxSize, comm=None, root=0, **kwargs):
        if comm.rank == root:
            array = numpy.array(array)
            if array.dtype.kind == 'c':
                # transform to real for the correct shape
                array = numpy.fft.irfftn(array)
                array[...] *= numpy.prod(array.shape)
            shape = array.shape
            dtype = array.dtype
        else:
            array, dtype, shape = [None] * 3

        dtype = comm.bcast(dtype, root=root)
        shape = comm.bcast(shape, root=root)

        assert len(shape) in (2, 3)

        Nmesh = shape

        empty = numpy.empty((0,), dtype)

        MeshSource.__init__(self, comm, Nmesh, BoxSize, empty.real.dtype)

        self.field = self.pm.create(mode='real')

        if comm.rank != root:
            array = empty # ignore data from other ranks.
        else:
            array = array.ravel()

        # fill the field with the array
        self.field.unravel(array)

    def to_real_field(self):
        if isinstance(self.field, RealField):
            return self.field.copy()
        else:
            return NotImplemented



from __future__ import print_function

import numpy
import logging
from mpi4py import MPI
from nbodykit.source.catalog import ArrayCatalog
from nbodykit.utils import split_size_3d

class FOF(object):
    """
    A friends-of-friends halo finder that computes the label for
    each particle, denoting which halo it belongs to.

    Friends-of-friends was first used by Davis et al 1985 to define
    halos in hierachical structure formation of cosmological simulations.
    The algorithm is also known as DBSCAN in computer science.
    The subroutine here implements a parallel version of the FOF.

    The underlying local FOF algorithm is from :mod:`kdcount.cluster`,
    which is an adaptation of the implementation in Volker Springel's
    Gadget and Martin White's PM.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~FOF.run` for the attributes storing the results.

    For returning a CatalogSource of the FOF halos, see :func:`find_features`
    and for computing a halo catalog with added analytic information for
    a specific redshift and cosmology, see :func:`to_halos`.

    Parameters
    ----------
    source : CatalogSource
        the source to run the FOF algorithm on; must support 'Position'
    linking_length : float
        the linking length, either in absolute units, or relative
        to the mean particle separation
    nmin : int
        halo with fewer particles are ignored
    absolute : bool, optional
        If `True`, the linking length is in absolute units, otherwise it is
        relative to the mean particle separation; default is `False`
    """
    logger = logging.getLogger('FOF')

    def __init__(self, source, linking_length, nmin, absolute=False, periodic=True, domain_factor=1):

        self.comm = source.comm
        self._source = source

        if 'Position' not in source:
            raise ValueError("cannot compute FOF without 'Position' column")

        self.attrs = {}
        self.attrs['linking_length'] = linking_length
        self.attrs['nmin'] = nmin
        self.attrs['absolute'] = absolute
        self.attrs['periodic'] = periodic
        self.attrs['domain_factor'] = domain_factor

        if periodic and 'BoxSize' not in source.attrs:
            raise ValueError("Periodic FOF requires BoxSize in .attrs['BoxSize']")

        # linking length relative to mean separation
        if not absolute:
            mean_separation = pow(numpy.prod(source.attrs['BoxSize']) / source.csize, 1.0 / len(source.attrs['Nmesh']))
            linking_length *= mean_separation
        self._linking_length = linking_length

        # and run
        self.run()

    def run(self):
        """
        Run the FOF algorithm. This function returns nothing, but does
        attach several attributes to the class instance:

        - attr:`labels`
        - :attr:`max_labels`

        .. note::
            The :attr:`labels` array is scattered evenly across all ranks.

        Attributes
        ----------
        labels : array_like, length: :attr:`size`
            an array the label that specifies which FOF halo each particle
            belongs to
        max_label : int
            the maximum label across all ranks; this represents the total
            number of FOF halos found
        """
        # run the FOF
        minid = fof(self._source, self._linking_length, self.comm, self.attrs['periodic'], self.attrs['domain_factor'], self.logger)

        # the sorted labels
        self.labels = _assign_labels(minid, comm=self.comm, thresh=self.attrs['nmin'])
        self.max_label = self.comm.allgather(self.labels.max())

    def find_features(self, peakcolumn=None):
        """
        Based on the particle labels, identify the groups, and return
        the center-of-mass ``CMPosition``, ``CMVelocity``, and Length of each
        feature.

        If a ``peakcolumn`` is given, the ``PeakPosition`` and ``PeakVelocity``
        is also calculated for the particle at the peak value of the column.

        Data is scattered evenly across all ranks.

        Returns
        -------
        :class:`~nbodykit.source.catalog.array.ArrayCatalog` :
            a source holding the ('CMPosition', 'CMVelocity', 'Length')
            of each feature; optionaly, ``PeakPosition``, ``PeakVelocity`` are
            also included if ``peakcolumn`` is not None
        """
        # the center-of-mass (Position, Velocity, Length)
        halos = fof_catalog(self._source, self.labels, self.comm, peakcolumn=peakcolumn, periodic=self.attrs['periodic'])
        attrs = self._source.attrs.copy()
        attrs.update(self.attrs)
        return ArrayCatalog(halos, comm=self.comm, **attrs)

    def to_halos(self, particle_mass, cosmo, redshift, mdef='vir',
                    posdef='cm', peakcolumn='Density'):
        """
        Return a :class:`~nbodykit.source.catalog.halos.HaloCatalog`, holding
        the center-of-mass position and velocity of each FOF halo, as well as
        the properly scaled mass, for a given cosmology and redshift.

        The returned catalog also has default analytic prescriptions for
        halo radius and concentration.

        The data is scattered evenly across all ranks.

        Parameters
        ----------
        particle_mass : float
            the particle mass, which is used to convert the number of particles
            in each halo to a total mass
        cosmo : :class:`nbodykit.cosmology.core.Cosmology`
            the cosmology of the catalog
        redshift : float
            the redshift of the catalog
        mdef : str, optional
            string specifying mass definition, used for computing default
            halo radii and concentration; should be 'vir' or 'XXXc' or
            'XXXm' where 'XXX' is an int specifying the overdensity
        posdef : str, optional
            position, can be cm (center of mass) or peak (particle with maximum value
            on a column)
        peakcolumn : str , optional
            when posdef is 'peak', this is the column in source for identifying
            particles at the peak for the position and velocity.

        Returns
        -------
        :class:`~nbodykit.source.catalog.halos.HaloCatalog`
            a HaloCatalog at the specified cosmology and redshift
        """
        from nbodykit.source.catalog.halos import HaloCatalog

        assert posdef in ['cm', 'peak'], "``posdef`` should be 'cm' or 'peak'"

        # meta-data
        attrs = self._source.attrs.copy()
        attrs.update(self.attrs)
        attrs['particle_mass'] = particle_mass

        if posdef == 'cm':
            # using the center-of-mass (Position, Velocity, Length) for each halo
            # not needing a column for peaks.
            peakcolumn = None
        else:
            pass
        data = fof_catalog(self._source, self.labels, self.comm, peakcolumn=peakcolumn, periodic=self.attrs['periodic'])
        data = data[data['Length'] > 0]
        halos = ArrayCatalog(data, comm=self.comm, **attrs)
        if posdef == 'cm':
            halos['Position'] = halos['CMPosition']
            halos['Velocity'] = halos['CMVelocity']
        elif posdef == 'peak':
            halos['Position'] = halos['PeakPosition']
            halos['Velocity'] = halos['PeakVelocity']
        # add the halo mass column
        halos['Mass'] = particle_mass * halos['Length']

        coldefs = {'mass':'Mass', 'velocity':'Velocity', 'position':'Position'}
        return HaloCatalog(halos, cosmo, redshift, mdef=mdef, **coldefs)

def _assign_labels(minid, comm, thresh):
    """
    Convert minid to sequential labels starting from 0.

    This routine is used to assign halo label to particles with
    the same minid.
    Halos with less than thresh particles are reclassified to 0.

    Parameters
    ----------
    minid : array_like, ('i8')
        The minimum particle id of the halo. All particles of a halo
        have the same minid
    comm : py:class:`MPI.Comm`
        communicator. since this is a collective operation
    thresh : int
        halo with less than thresh particles are merged into halo 0

    Returns
    -------
    labels : array_like ('i8')
        The new labels of particles. Note that this is ordered
        by the size of halo, with the exception 0 represents all
        particles that are in halos that contain less than thresh particles.

    """
    from mpi4py import MPI

    dtype = numpy.dtype([
            ('origind', 'u8'),
            ('fofid', 'u8'),
            ])
    data = numpy.empty(len(minid), dtype=dtype)
    # assign origind for recovery of ordering, since
    # we need to work in sorted fofid
    data['fofid'] = minid
    data['origind'] = numpy.arange(len(data), dtype='u4')
    data['origind'] += sum(comm.allgather(len(data))[:comm.rank]) \

    data = DistributedArray(data, comm)

    # first attempt is to assign fofid for each group
    data.sort('fofid')
    label = data['fofid'].unique_labels()

    N = label.bincount()

    # now eliminate those with less than thresh particles
    small = N.local <= thresh

    Nlocal = label.bincount(local=True)
    # mask == True for particles in small halos
    mask = numpy.repeat(small, Nlocal)

    # globally shift halo id by one
    label.local += 1
    label.local[mask] = 0

    data['fofid'].local[:] = label.local[:]
    del label

    data.sort('fofid')

    data['fofid'].local[:] = data['fofid'].unique_labels().local[:]

    data.sort('origind')

    label = data['fofid'].local.view('i8').copy()

    del data

    Nhalo0 = max(comm.allgather(label.max())) + 1
    Nlocal = numpy.bincount(label, minlength=Nhalo0)
    comm.Allreduce(MPI.IN_PLACE, Nlocal, op=MPI.SUM)

    # sort the labels by halo size
    arg = Nlocal[1:].argsort()[::-1] + 1
    P = numpy.arange(Nhalo0, dtype='i4')
    P[arg] = numpy.arange(len(arg), dtype='i4') + 1
    label = P[label]

    return label

def _fof_local(layout, pos, boxsize, ll, comm):
    from kdcount import cluster

    N = len(pos)

    pos = layout.exchange(pos)
    if boxsize is not None:
        pos %= boxsize
    data = cluster.dataset(pos, boxsize=boxsize)
    fof = cluster.fof(data, linking_length=ll, np=0)
    labels = fof.labels
    del fof

    PID = numpy.arange(N, dtype='intp')
    PID += sum(comm.allgather(N)[:comm.rank])

    PID = layout.exchange(PID)
    # initialize global labels
    minid = equiv_class(labels, PID, op=numpy.fmin)[labels]

    return minid

def _fof_merge(layout, minid, comm):
    # generate global halo id

    while True:
        # merge, if a particle belongs to several ranks
        # use the global label of the minimal
        minid_new = layout.gather(minid, mode=numpy.fmin)
        minid_new = layout.exchange(minid_new)

        # on my rank, these particles have been merged
        merged = minid_new != minid
        # if no rank has merged any, we are done
        # gl is the global label (albeit with some holes)
        total = comm.allreduce(merged.sum())

        if total == 0:
            del minid_new
            break
        old = minid[merged]
        new = minid_new[merged]
        arg = old.argsort()
        new = new[arg]
        old = old[arg]
        replacesorted(minid, old, new, out=minid)

    minid = layout.gather(minid, mode=numpy.fmin)
    return minid

def fof(source, linking_length, comm, periodic, domain_factor, logger):
    """
    Run Friends-of-friends halo finder.

    Friends-of-friends was first used by Davis et al 1985 to define
    halos in hierachical structure formation of cosmological simulations.
    The algorithm is also known as DBSCAN in computer science.
    The subroutine here implements a parallel version of the FOF.

    The underlying local FOF algorithm is from `kdcount.cluster`,
    which is an adaptation of the implementation in Volker Springel's
    Gadget and Martin White's PM. It could have been done faster.

    Parameters
    ----------
    source: CatalogSource
        the input source of particles; must support 'Position' column;
        ``source.attrs['BoxSize']`` is also used
    linking_length: float
        linking length in data units. (Usually Mpc/h).
    comm: MPI.Comm
        The mpi communicator.

    Returns
    -------
    minid: array_like
        A unique label of each position. The label is not ranged from 0.
    """
    from pmesh.domain import GridND

    np = split_size_3d(comm.size)
    nd = np * domain_factor

    if periodic:
        BoxSize = source.attrs.get('BoxSize', None)
        if BoxSize is None:
            raise ValueError("cannot compute FOF clustering of source without 'BoxSize' in ``attrs`` dict")
        if numpy.isscalar(BoxSize):
            BoxSize = [BoxSize, BoxSize, BoxSize]

        left = [0, 0, 0]
        right = BoxSize
    else:
        BoxSize = None
        left = numpy.min(comm.allgather(source['Position'].min(axis=0).compute()), axis=0)
        right = numpy.max(comm.allgather(source['Position'].max(axis=0).compute()), axis=0)

    grid = [
        numpy.linspace(left[0], right[0], nd[0] + 1, endpoint=True),
        numpy.linspace(left[1], right[1], nd[1] + 1, endpoint=True),
        numpy.linspace(left[2], right[2], nd[2] + 1, endpoint=True),
    ]
    domain = GridND(grid, comm=comm, periodic=periodic)

    Position = source.compute(source['Position'])
    np = comm.allgather(len(Position))
    if comm.rank == 0:
        logger.info("Number of particles max/min = %d / %d before spatial decomposition" % (max(np), min(np)))

    # balance the load
    domain.loadbalance(domain.load(Position))

    layout = domain.decompose(Position, smoothing=linking_length * 1)

    np = comm.allgather(layout.newlength)
    if comm.rank == 0:
        logger.info("Number of particles max/min = %d / %d after spatial decomposition" % (max(np), min(np)))

    comm.barrier()
    minid = _fof_local(layout, Position, BoxSize, linking_length, comm)

    comm.barrier()
    minid = _fof_merge(layout, minid, comm)

    return minid

def fof_find_peaks(source, label, comm,
                position='Position', column='Density'):
    """
    Find position of the peak (maximum) from a given column for a fof result.
    """
    Nhalo0 = max(comm.allgather(label.max())) + 1

    N = numpy.bincount(label, minlength=Nhalo0)
    comm.Allreduce(MPI.IN_PLACE, N, op=MPI.SUM)

    return hpos

def fof_catalog(source, label, comm,
                position='Position', velocity='Velocity', initposition='InitialPosition',
                peakcolumn=None, periodic=True):
    """
    Catalog of FOF groups based on label from a parent source

    This is a collective operation -- the returned halo catalog will be
    equally distributed across all ranks

    Notes
    -----
    This computes the center-of-mass position and velocity in the same
    units as the corresponding columns ``source``

    Parameters
    ----------
    source: CatalogSource
        the parent source of particles from which the center-of-mass
        position and velocity are computed for each halo
    label : array_like
        the label for each particle that identifies which halo it
        belongs to
    comm: MPI.Comm
        the mpi communicator. Must agree with the datasource
    position : str, optional
        the column name specifying the position
    velocity : str, optional
        the column name specifying the velocity
    initposition : str, optional
        the column name specifying the initial position; this is only
        computed if available
    peakcolumn : str, optional
        if not None, find PeakPostion and PeakVelocity based on the
        value of peakcolumn

    Returns
    -------
    catalog: array_like
        A 1-d array of type 'Position', 'Velocity', 'Length'.
        The center mass position and velocity of the FOF halo, and
        Length is the number of particles in a halo. The catalog is
        sorted such that the most massive halo is first. ``catalog[0]``
        does not correspond to any halo.
    """
    from nbodykit.utils import ScatterArray

    # make sure all of the columns are there
    for col in [position, velocity]:
        if col not in source:
            raise ValueError("the column '%s' is missing from parent source; cannot compute halos" %col)

    dtype=[('CMPosition', ('f4', 3)),('CMVelocity', ('f4', 3)),('Length', 'i4')]
    N = count(label, comm=comm)

    if periodic:
        # make sure BoxSize is there
        boxsize = source.attrs.get('BoxSize', None)
        if boxsize is None:
            raise ValueError("cannot compute halo catalog from source without 'BoxSize' in ``attrs`` dict")
    else:
        boxsize = None

    # center of mass position
    hpos = centerofmass(label, source.compute(source[position]), boxsize=boxsize, comm=comm)

    # center of mass velocity
    hvel = centerofmass(label, source.compute(source[velocity]), boxsize=None, comm=comm)

    # center of mass initial position
    if initposition in source:
        dtype.append(('InitialPosition', ('f4', 3)))
        hpos_init = centerofmass(label, source.compute(source[initposition]), boxsize=boxsize, comm=comm)

    if peakcolumn is not None:
        assert peakcolumn in source

        dtype.append(('PeakPosition', ('f4', 3)))
        dtype.append(('PeakVelocity', ('f4', 3)))

        density = source[peakcolumn].compute()
        dmax = equiv_class(label, density, op=numpy.fmax, dense_labels=True, minlength=len(N), identity=-numpy.inf)
        comm.Allreduce(MPI.IN_PLACE, dmax, op=MPI.MAX)
        # remove any non-peak particle from the labels
        label1 = label * (density >= dmax[label])

        # compute the center of mass on the new labels
        ppos = centerofmass(label1, source.compute(source[position]), boxsize=boxsize, comm=comm)
        pvel = centerofmass(label1, source.compute(source[velocity]), boxsize=None, comm=comm)

    dtype = numpy.dtype(dtype)
    if comm.rank == 0:
        catalog = numpy.empty(shape=len(N), dtype=dtype)

        catalog['CMPosition'] = hpos
        catalog['CMVelocity'] = hvel
        catalog['Length'] = N
        catalog['Length'][0] = 0
        if 'InitialPosition' in dtype.names:
            catalog['InitialPosition'] = hpos_init

        if peakcolumn is not None:
            catalog['PeakPosition'] = ppos
            catalog['PeakVelocity'] = pvel
    else:
        catalog = None

    return ScatterArray(catalog, comm, root=0)

# -----------------------
# Helpers
# -----------------------
def equiv_class(labels, values, op, dense_labels=False, identity=None, minlength=None):
    """
    apply operation to equivalent classes by label, on values

    Parameters
    ----------
    labels : array_like
        the label of objects, starting from 0.
    values : array_like
        the values of objects (len(labels), ...)
    op : :py:class:`numpy.ufunc`
        the operation to apply
    dense_labels : boolean
        If the labels are already dense (from 0 to Nobjects - 1)
        If False, :py:meth:`numpy.unique` is used to convert
        the labels internally

    Returns
    -------
    result :
        the value of each equivalent class

    Examples
    --------
    >>> x = numpy.arange(10)
    >>> print equiv_class(x, x, numpy.fmin, dense_labels=True)
    [0 1 2 3 4 5 6 7 8 9]

    >>> x = numpy.arange(10)
    >>> v = numpy.arange(20).reshape(10, 2)
    >>> x[1] = 0
    >>> print equiv_class(x, 1.0 * v, numpy.fmin, dense_labels=True, identity=numpy.inf)
    [[  0.   1.]
     [ inf  inf]
     [  4.   5.]
     [  6.   7.]
     [  8.   9.]
     [ 10.  11.]
     [ 12.  13.]
     [ 14.  15.]
     [ 16.  17.]
     [ 18.  19.]]

    """
    # dense labels
    if not dense_labels:
        junk, labels = numpy.unique(labels, return_inverse=True)
        del junk
    N = numpy.bincount(labels)
    offsets = numpy.concatenate([[0], N.cumsum()], axis=0)[:-1]
    arg = labels.argsort()
    if identity is None: identity = op.identity
    if minlength is None:
        minlength = len(N)

    # work around numpy dtype reference counting bugs
    # be a simple man and never steal anything.

    dtype = numpy.dtype((values.dtype, values.shape[1:]))

    result = numpy.empty(minlength, dtype=dtype)
    result[:len(N)] = op.reduceat(values[arg], offsets)

    if (N == 0).any():
        result[:len(N)][N == 0] = identity

    if len(N) < minlength:
        result[len(N):] = identity

    return result

def replacesorted(arr, sorted, b, out=None):
    """
    replace a with corresponding b in arr

    Parameters
    ----------
    arr : array_like
        input array
    sorted   : array_like
        sorted

    b   : array_like

    out : array_like,
        output array
    Result
    ------
    newarr  : array_like
        arr with a replaced by corresponding b

    Examples
    --------
    >>> print replacesorted(numpy.arange(10), numpy.arange(5), numpy.ones(5))
    [1 1 1 1 1 5 6 7 8 9]

    """
    if out is None:
        out = arr.copy()
    if len(sorted) == 0:
        return out
    ind = sorted.searchsorted(arr)
    ind.clip(0, len(sorted) - 1, out=ind)
    arr = numpy.array(arr)
    found = sorted[ind] == arr
    out[found] = b[ind[found]]
    return out

import mpsort
class DistributedArray(object):
    """
    Distributed Array Object

    A distributed array is striped along ranks

    Attributes
    ----------
    comm : :py:class:`mpi4py.MPI.Comm`
        the communicator

    local : array_like
        the local data

    """
    def __init__(self, local, comm):
        self.local = local
        self.comm = comm
        self.topology = LinearTopology(local, comm)

    def sort(self, orderby=None):
        """
        Sort array globally by key orderby.

        Due to a limitation of mpsort, self[orderby] must be u8.

        """
        mpsort.sort(self.local, orderby, comm=self.comm)

    def __getitem__(self, key):
        return DistributedArray(self.local[key], self.comm)

    def unique_labels(self):
        """
        Assign unique labels to sorted local.

        .. warning ::

            local data must be sorted, and of simple type. (numpy.unique)

        Returns
        -------
        label   :  :py:class:`DistributedArray`
            the new labels, starting from 0

        """
        prev, next = self.topology.prev(), self.topology.next()

        junk, label = numpy.unique(self.local, return_inverse=True)

        if len(label) == 0:
            # work around numpy bug (<=1.13.3) when label is empty it
            # spits out booleans?? booleans!!
            # this causes issues when type cast rules in numpy
            # are tighten up.
            label = numpy.int64(label)

        if len(self.local) == 0:
            Nunique = 0
        else:
            # watch out: this is to make sure after shifting first
            # labels on the next rank is the same as my last label
            # when there is a spill-over.
            if next == self.local[-1]:
                Nunique = len(junk) - 1
            else:
                Nunique = len(junk)

        label += sum(self.comm.allgather(Nunique)[:self.comm.rank])
        return DistributedArray(label, self.comm)

    def bincount(self, local=False):
        """
        Assign count numbers from sorted local data.

        .. warning ::

            local data must be sorted, and of integer type. (numpy.bincount)

        Parameters
        ----------
        local : boolean
            if local is True, only count the local array.

        Returns
        -------
        N :  :py:class:`DistributedArray`
            distributed counts array. If items of the same value spans other
            chunks of array, they are added to N as well.

        Examples
        --------
        if the local array is [ (0, 0), (0, 1)],
        Then the counts array is [ (3, ), (3, 1)]
        """
        prev = self.topology.prev()
        if prev is not EmptyRank:
            offset = prev
            if len(self.local) > 0:
                if prev != self.local[0]:
                    offset = self.local[0]
        else:
            offset = 0

        N = numpy.bincount(self.local - offset)

        if local:
            return N

        heads = self.topology.heads()
        tails = self.topology.tails()

        distN = DistributedArray(N, self.comm)
        headsN, tailsN = distN.topology.heads(), distN.topology.tails()

        if len(N) > 0:
            for i in reversed(range(self.comm.rank)):
                if tails[i] == self.local[0]:
                    N[0] += tailsN[i]
            for i in range(self.comm.rank + 1, self.comm.size):
                if heads[i] == self.local[-1]:
                    N[-1] += headsN[i]

        return DistributedArray(N, self.comm)

class EmptyRankType(object):
    def __repr__(self):
        return "EmptyRank"
EmptyRank = EmptyRankType()

class LinearTopology(object):
    """ Helper object for the topology of a distributed array
    """
    def __init__(self, local, comm):
        self.local = local
        self.comm = comm

    def heads(self):
        """
        The first items on each rank.

        Returns
        -------
        heads : list
            a list of first items, EmptyRank is used for empty ranks
        """

        head = EmptyRank
        if len(self.local) > 0:
            head = self.local[0]

        return self.comm.allgather(head)

    def tails(self):
        """
        The last items on each rank.

        Returns
        -------
        tails: list
            a list of last items, EmptyRank is used for empty ranks
        """
        tail = EmptyRank
        if len(self.local) > 0:
            tail = self.local[-1]

        return self.comm.allgather(tail)

    def prev(self):
        """
        The item before the local data.

        This method fetches the last item before the local data.
        If the rank before is empty, the rank before is used.

        If no item is before this rank, EmptyRank is returned

        Returns
        -------
        prev : scalar
            Item before local data, or EmptyRank if all ranks before this rank is empty.

        """

        tails = [EmptyRank]
        oldtail = EmptyRank
        for tail in self.tails():
            if tail is EmptyRank:
                tails.append(oldtail)
            else:
                tails.append(tail)
                oldtail = tail
        prev = tails[self.comm.rank]
        return prev

    def next(self):
        """
        The item after the local data.

        This method the first item after the local data.
        If the rank after current rank is empty,
        item after that rank is used.

        If no item is after local data, EmptyRank is returned.

        Returns
        -------
        next : scalar
            Item after local data, or EmptyRank if all ranks after this rank is empty.

        """
        heads = []
        oldhead = EmptyRank
        for head in self.heads():
            if head is EmptyRank:
                heads.append(oldhead)
            else:
                heads.append(head)
                oldhead = head
        heads.append(EmptyRank)

        next = heads[self.comm.rank + 1]
        return next

def centerofmass(label, pos, boxsize, comm=MPI.COMM_WORLD):
    """
    Calulate the center of mass of particles of the same label.

    The center of mass is defined as the mean of positions of particles,
    but care has to be taken regarding to the periodic boundary.

    This is a collective operation, and after the call, all ranks
    will have the position of halos.

    Parameters
    ----------
    label : array_like (integers)
        Halo label of particles, >=0
    pos   : array_like (float, 3)
        position of particles.
    boxsize : float or None
        size of the periodic box, or None if no periodic boundary is assumed.
    comm : :py:class:`MPI.Comm`
        communicator for the collective operation.

    Returns
    -------
    hpos : array_like (float, 3)
        the center of mass position of the halos.

    """
    Nhalo0 = max(comm.allgather(label.max())) + 1

    N = numpy.bincount(label, minlength=Nhalo0)
    comm.Allreduce(MPI.IN_PLACE, N, op=MPI.SUM)

    if boxsize is not None:
        posmin = equiv_class(label, pos, op=numpy.fmin, dense_labels=True, identity=numpy.inf,
                        minlength=len(N))
        comm.Allreduce(MPI.IN_PLACE, posmin, op=MPI.MIN)
        dpos = pos - posmin[label]
        for i in range(dpos.shape[-1]):
            bhalf = boxsize[i] * 0.5
            dpos[..., i][dpos[..., i] < -bhalf] += boxsize[i]
            dpos[..., i][dpos[..., i] >= bhalf] -= boxsize[i]
    else:
        dpos = pos
    dpos = equiv_class(label, dpos, op=numpy.add, dense_labels=True, minlength=len(N))

    comm.Allreduce(MPI.IN_PLACE, dpos, op=MPI.SUM)
    dpos /= N[:, None]

    if boxsize is not None:
        hpos = posmin + dpos
        hpos %= boxsize
    else:
        hpos = dpos
    return hpos

def count(label, comm=MPI.COMM_WORLD):
    """
    Count the number of particles of the same label.

    This is a collective operation, and after the call, all ranks
    will have the particle count.

    Parameters
    ----------
    label : array_like (integers)
        Halo label of particles, >=0
    comm : :py:class:`MPI.Comm`
        communicator for the collective operation.

    Returns
    -------
    count : array_like
        the count of number of particles in each halo

    """
    Nhalo0 = max(comm.allgather(label.max())) + 1

    N = numpy.bincount(label, minlength=Nhalo0)
    comm.Allreduce(MPI.IN_PLACE, N, op=MPI.SUM)

    return N



import os
import numpy
import logging

from nbodykit import CurrentMPIComm
from nbodykit.binned_statistic import BinnedStatistic
from nbodykit.meshtools import SlabIterator
from nbodykit.base.catalog import CatalogSourceBase
from nbodykit.base.mesh import MeshSource

from .fftpower import FFTBase
from .fftpower import project_to_basis
from .fftpower import _find_unique_edges

class FFTCorr(FFTBase):
    r"""
    Algorithm to compute the 1d or 2d correlation and/or multipoles
    in a periodic box, using a Fast Fourier Transform (FFT).

    This computes the power spectrum as the square of the Fourier modes of the
    density field, which are computed via a FFT. Then it is transformed back
    to obtain the correlation function.

    Results are computed when the object is inititalized. See the documenation
    of :func:`~FFTCorr.run` for the attributes storing the results.

    .. note::
        This is very similar to :class:`~nbodykit.algorithms.fftpower.FFTPower`. 

    Parameters
    ----------
    first : CatalogSource, MeshSource
        the source for the first field; if a CatalogSource is provided, it
        is automatically converted to MeshSource using the default painting
        parameters (via :func:`~nbodykit.base.catalogmesh.CatalogMesh.to_mesh`)
    mode : {'1d', '2d'}
        compute either 1d or 2d power spectra
    Nmesh : int, optional
        the number of cells per side in the particle mesh used to paint the source
    BoxSize : int, 3-vector, optional
        the size of the box
    second : CatalogSource, MeshSource, optional
        the second source for cross-correlations
    los : array_like , optional
        the direction to use as the line-of-sight; must be a unit vector
    Nmu : int, optional
        the number of mu bins to use from :math:`\mu=[0,1]`;
        if `mode = 1d`, then ``Nmu`` is set to 1
    dr : float, optional
        the linear spacing of ``r`` bins to use; if not provided, the
        fundamental mode  of the box is used; if `dr=0`, the bins are tight, such
        that each bin has a unique r value.
    rmin : float, optional
        the lower edge of the first ``r`` bin to use
    poles : list of int, optional
        a list of multipole numbers ``ell`` to compute :math:`\xi_\ell(r)`
        from :math:`\xi(r,\mu)`
    """
    logger = logging.getLogger('FFTCorr')

    def __init__(self, first, mode, Nmesh=None, BoxSize=None, second=None,
                    los=[0, 0, 1], Nmu=5, dr=None, rmin=0., poles=[]):

        # mode is either '1d' or '2d'
        if mode not in ['1d', '2d']:
            raise ValueError("`mode` should be either '1d' or '2d'")

        if poles is None:
            poles = []

        # check los
        if numpy.isscalar(los) or len(los) != 3:
            raise ValueError("line-of-sight ``los`` should be vector with length 3")
        if not numpy.allclose(numpy.einsum('i,i', los, los), 1.0, rtol=1e-5):
            raise ValueError("line-of-sight ``los`` must be a unit vector")

        FFTBase.__init__(self, first, second, Nmesh, BoxSize)

        # save meta-data
        self.attrs['mode'] = mode
        self.attrs['los'] = los
        self.attrs['Nmu'] = Nmu
        self.attrs['poles'] = poles

        if dr is None:
            dr = self.attrs['BoxSize'].min() / self.attrs['Nmesh'].max()

        self.attrs['dr'] = dr
        self.attrs['rmin'] = rmin

        self.corr, self.poles = self.run()

        # compatability
        self.attrs.update(self.corr.attrs)

    def run(self):
        r"""
        Compute the correlation function in a periodic box, using FFTs.

        returns
        -------
        corr : :class:`~nbodykit.binned_statistic.BinnedStatistic`
            a BinnedStatistic object that holds the measured :math:`\xi(r)` or
            :math:`\xi(r,\mu)`. It stores the following variables:

            - r :
                the mean value for each ``r`` bin
            - mu : ``mode=2d`` only
                the mean value for each ``mu`` bin
            - corr :
                real array storing the correlation function
            - modes :
                the number of modes averaged together in each bin

        poles : :class:`~nbodykit.binned_statistic.BinnedStatistic` or ``None``
            a BinnedStatistic object to hold the multipole results
            :math:`\xi_\ell(r)`; if no multipoles were requested by the user,
            this is ``None``. It stores the following variables:

            - r :
                the mean value for each ``r`` bin
            - power_L :
                complex array storing the real and imaginary components for
                the :math:`\ell=L` multipole
            - modes :
                the number of modes averaged together in each bin

        corr.attrs, poles.attrs : dict
            dictionary of meta-data; in addition to storing the input parameters,
            it includes the following fields computed during the algorithm
            execution:

            - shotnoise : float
                the power Poisson shot noise, equal to :math:`V/N`, where
                :math:`V` is the volume of the box and `N` is the total
                number of objects; if a cross-correlation is computed, this
                will be equal to zero
            - N1 : int
                the total number of objects in the first source
            - N2 : int
                the total number of objects in the second source
        """

        # only need one mu bin if 1d case is requested
        if self.attrs['mode'] == "1d": self.attrs['Nmu'] = 1

        # measure the 3D power (y3d is a ComplexField)
        y3d, attrs = self._compute_3d_power(self.first, self.second)

        # measure the 3D correlation (y3d is a RealField)
        y3d = y3d.c2r(out=Ellipsis)

        # correlation is dimensionless
        # Note that L^3 cancels with dk^3.
        y3d[...] *= 1.0 / y3d.BoxSize.prod()

        # binning in k out to the minimum nyquist frequency
        # (accounting for possibly anisotropic box)
        dr = self.attrs['dr']
        rmin = self.attrs['rmin']
        if dr > 0:
            redges = numpy.arange(rmin, 0.5 * y3d.BoxSize.min() + dr/2, dr)
            rcenters = None
        else:
            redges, rcenters = _find_unique_edges(y3d.x, y3d.BoxSize / y3d.Nmesh, self.comm)

        # project on to the desired basis
        muedges = numpy.linspace(0, 1, self.attrs['Nmu']+1, endpoint=True)
        edges = [redges, muedges]
        coords = [rcenters, None]
        result, pole_result = project_to_basis(y3d, edges,
                                               poles=self.attrs['poles'],
                                               los=self.attrs['los'])

        # format the corr results into structured array
        if self.attrs['mode'] == "1d":
            cols = ['r', 'corr', 'modes']
            icols = [0, 2, 3]
            edges = edges[0:1]
            coords = coords[0:1]
        else:
            cols = ['r', 'mu', 'corr', 'modes']
            icols = [0, 1, 2, 3]

        # corr results as a structured array
        dtype = numpy.dtype([(name, result[icol].dtype.str) for icol,name in zip(icols,cols)])
        corr = numpy.squeeze(numpy.empty(result[0].shape, dtype=dtype))
        for icol, col in zip(icols, cols):
            corr[col][:] = numpy.squeeze(result[icol])

        # multipole results as a structured array
        poles = None
        if pole_result is not None:
            r, poles, N = pole_result
            cols = ['r'] + ['corr_%d' %l for l in self.attrs['poles']] + ['modes']
            result = [r] + [pole for pole in poles] + [N]

            dtype = numpy.dtype([(name, result[icol].dtype.str) for icol,name in enumerate(cols)])
            poles = numpy.empty(result[0].shape, dtype=dtype)
            for icol, col in enumerate(cols):
                poles[col][:] = result[icol]

        # set all the necessary results
        return self._make_datasets(edges, poles, corr, coords, attrs)

    def __getstate__(self):
        state = dict(
                    corr=self.corr.__getstate__(),
                    poles=self.poles.__getstate__() if self.poles is not None else None,
                    attrs=self.attrs)
        return state

    def __setstate__(self, state):
        self.attrs = state['attrs']
        self.corr = BinnedStatistic.from_state(state['corr'])
        if state['poles'] is not None:
            self.poles = BinnedStatistic.from_state(state['poles'])

    def _make_datasets(self, edges, poles, corr, coords, attrs):

        if self.attrs['mode'] == '1d':
            corr = BinnedStatistic(['r'], edges, corr, fields_to_sum=['modes'], coords=coords, **attrs)
        else:
            corr = BinnedStatistic(['r', 'mu'], edges, corr, fields_to_sum=['modes'], coords=coords, **attrs)
        if poles is not None:
            poles = BinnedStatistic(['r'], [corr.edges['r']], poles, fields_to_sum=['modes'], coords=coords, **attrs)

        return corr, poles
